import json, sys, os

# Set the project directory to the parent directory of the current file
project_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if project_dir not in sys.path: 
    sys.path.append(project_dir)

# Attempt to import the MediLink_ConfigLoader module, falling back to an alternative import if necessary
try: 
    import MediLink_ConfigLoader
except ImportError: 
    from MediLink import MediLink_ConfigLoader

# Attempt to import the fetch_payer_name_from_api function from MediLink_API_v3, with a fallback
try: 
    from MediLink_API_v3 import fetch_payer_name_from_api
except ImportError: 
    from MediLink import MediLink_API_v3
    fetch_payer_name_from_api = MediLink_API_v3.fetch_payer_name_from_api

# Attempt to import the MediBot_Preprocessor_lib module, with a fallback
try: 
    from MediBot import MediBot_Preprocessor_lib
except ImportError: 
    import MediBot_Preprocessor_lib

"""
# TODO This has a bunch of issues that need to be fixed. Notice, the log has repetitive calls to the API that are redundant.

PS G:\My Drive\Codes\MediCafe> & C:/Python34/python.exe "g:/My Drive/Codes/MediCafe/MediBot/MediBot_Preprocessor.py" --update-crosswalk
Updating the crosswalk...
The 'payer_id' list is empty or missing. Would you like to initialize the crosswalk? (yes/no): yes
No payer found at AVAILITY for ID 60054M. Response: {'limit': 50, 'offset': 0, 'totalCount': 0, 'payers': [], 'count': 0}
All endpoints exhausted for Payer ID 60054M.
WARNING: Invalid Payer ID 60054M (Unknown).
Enter the correct Payer ID for replacement or type 'FORCE' to continue with the unresolved Payer ID: 60054
csv_replacements updated: '60054M' -> '60054'.
No payer found at AVAILITY for ID MCRFL. Response: {'limit': 50, 'offset': 0, 'totalCount': 0, 'payers': [], 'count': 0}
All endpoints exhausted for Payer ID MCRFL.
No payer found at AVAILITY for ID BCSFL. Response: {'limit': 50, 'offset': 0, 'totalCount': 0, 'payers': [], 'count': 0}
All endpoints exhausted for Payer ID BCSFL.
Payer ID '60054M' has been successfully replaced with '60054'.
No payer found at AVAILITY for ID MCRFL. Response: {'limit': 50, 'offset': 0, 'totalCount': 0, 'payers': [], 'count': 0}
All endpoints exhausted for Payer ID MCRFL.
WARNING: Invalid Payer ID MCRFL (Unknown).
Enter the correct Payer ID for replacement or type 'FORCE' to continue with the unresolved Payer ID: force
Payer ID 'MCRFL' has been marked as 'Unknown'.
No payer found at AVAILITY for ID BCSFL. Response: {'limit': 50, 'offset': 0, 'totalCount': 0, 'payers': [], 'count': 0}
All endpoints exhausted for Payer ID BCSFL.
WARNING: Invalid Payer ID BCSFL (Unknown).
Enter the correct Payer ID for replacement or type 'FORCE' to continue with the unresolved Payer ID: 00590
csv_replacements updated: 'BCSFL' -> '00590'.
Payer ID 'BCSFL' has been successfully replaced with '00590'.
Crosswalk initialized with mappings for 5 payers.

"""


def fetch_and_store_payer_name(client, payer_id, crosswalk, config):
    """
    Fetches the payer name for a given payer ID and stores it in the crosswalk.
    
    Args:
        payer_id (str): The ID of the payer to fetch.
        crosswalk (dict): The crosswalk dictionary to store the payer name.
        config (dict): Configuration settings for logging.

    Returns:
        bool: True if the payer name was fetched and stored successfully, False otherwise.
    """
    MediLink_ConfigLoader.log("Attempting to fetch payer name for Payer ID: {}".format(payer_id), config, level="DEBUG")
    try:
        # Fetch the payer name from the API
        payer_name = fetch_payer_name_from_api(client, payer_id, config, primary_endpoint=None)
        MediLink_ConfigLoader.log("Fetched payer name: {} for Payer ID: {}".format(payer_name, payer_id), config, level="DEBUG")
        
        # Ensure the 'payer_id' key exists in the crosswalk
        if 'payer_id' not in crosswalk: 
            crosswalk['payer_id'] = {}
            MediLink_ConfigLoader.log("Initialized 'payer_id' in crosswalk.", config, level="DEBUG")
        
        # Initialize the payer ID entry if it doesn't exist
        if payer_id not in crosswalk['payer_id']:
            crosswalk['payer_id'][payer_id] = {}  # Initialize the entry
            MediLink_ConfigLoader.log("Initialized entry for Payer ID: {}".format(payer_id), config, level="DEBUG")
        
        # Store the fetched payer name in the crosswalk
        crosswalk['payer_id'][payer_id]['name'] = payer_name
        message = "Payer ID {} ({}) fetched and stored successfully.".format(payer_id, payer_name)
        MediLink_ConfigLoader.log(message, config, level="INFO")
        print(message)
        return True
    except Exception as e:
        # Log any errors encountered during the fetching process
        MediLink_ConfigLoader.log("Failed to fetch name for Payer ID {}: {}".format(payer_id, e), config, level="WARNING")
        crosswalk['payer_id'][payer_id]['name'] = "Unknown"
        return False

def validate_and_correct_payer_ids(client, crosswalk, config, auto_correct=False):
    """
    Validates and corrects payer IDs in the crosswalk. If a payer ID is invalid, it prompts the user for correction.
    
    Args:
        crosswalk (dict): The crosswalk dictionary containing payer IDs.
        config (dict): Configuration settings for logging.
        auto_correct (bool): If True, automatically corrects invalid payer IDs to 'Unknown'.
    """
    processed_payer_ids = set()  # Track processed payer IDs
    payer_ids = list(crosswalk.get('payer_id', {}).keys())  # Static list to prevent modification issues
    
    for payer_id in payer_ids:
        if payer_id in processed_payer_ids:
            continue  # Skip already processed payer IDs
        
        # Validate the payer ID by fetching its name
        is_valid = fetch_and_store_payer_name(client, payer_id, crosswalk, config)
        
        if not is_valid:
            if auto_correct:
                # Automatically correct invalid payer IDs to 'Unknown'
                crosswalk['payer_id'][payer_id]['name'] = "Unknown"
                MediLink_ConfigLoader.log(
                    "Auto-corrected Payer ID {} to 'Unknown'.".format(payer_id),
                    config,
                    level="WARNING"
                )
                print("Auto-corrected Payer ID '{}' to 'Unknown'.".format(payer_id))
                processed_payer_ids.add(payer_id)
                continue
            
            # Prompt the user for a corrected payer ID
            current_name = crosswalk['payer_id'].get(payer_id, {}).get('name', 'Unknown')
            corrected_payer_id = input(
                "WARNING: Invalid Payer ID {} ({}).\n"
                "Enter the correct Payer ID for replacement or type 'FORCE' to continue with the unresolved Payer ID: ".format(
                    payer_id, current_name)
            ).strip()
            
            if corrected_payer_id.lower() == 'force':
                # Assign "Unknown" and log the action
                crosswalk['payer_id'][payer_id]['name'] = "Unknown"
                MediLink_ConfigLoader.log(
                    "User forced unresolved Payer ID {} to remain as 'Unknown'.".format(payer_id),
                    config,
                    level="WARNING"
                )
                print("Payer ID '{}' has been marked as 'Unknown'.".format(payer_id))
                processed_payer_ids.add(payer_id)
                continue
            
            if corrected_payer_id:
                # Validate the corrected payer ID
                if fetch_and_store_payer_name(client, corrected_payer_id, crosswalk, config):
                    # Replace the old payer ID with the corrected one in the crosswalk
                    success = update_crosswalk_with_corrected_payer_id(client, payer_id, corrected_payer_id, config, crosswalk)
                    if success:
                        print("Payer ID '{}' has been successfully replaced with '{}'.".format(
                            payer_id, corrected_payer_id))
                        processed_payer_ids.add(corrected_payer_id)
                else:
                    # Only set to "Unknown" if the corrected payer ID is not valid
                    crosswalk['payer_id'][corrected_payer_id] = {'name': "Unknown"}
                    MediLink_ConfigLoader.log(
                        "Failed to validate corrected Payer ID {}. Set to 'Unknown'.".format(corrected_payer_id),
                        config,
                        level="ERROR"
                    )
                    print("Payer ID '{}' has been added with name 'Unknown'.".format(corrected_payer_id))
                    processed_payer_ids.add(corrected_payer_id)
            else:
                MediLink_ConfigLoader.log(
                    "No correction provided for Payer ID {}. Skipping.".format(payer_id),
                    config,
                    level="WARNING"
                )
                print("No correction provided for Payer ID '{}'. Skipping.".format(payer_id))

def initialize_crosswalk_from_mapat(client, config, crosswalk):
    """
    Initializes the crosswalk from the MAPAT data source. Loads configuration and data sources, 
    validates payer IDs, and saves the crosswalk.
    
    Returns:
        dict: The payer ID mappings from the initialized crosswalk.
    """
    # Ensure full configuration and crosswalk are loaded
    config, crosswalk = ensure_full_config_loaded(config, crosswalk)
    
    try:
        # Load data sources for patient and payer IDs
        patient_id_to_insurance_id, payer_id_to_patient_ids = MediBot_Preprocessor_lib.load_data_sources(config, crosswalk)
    except ValueError as e:
        print(e)
        sys.exit(1)
    
    # Map payer IDs to insurance IDs
    payer_id_to_details = MediBot_Preprocessor_lib.map_payer_ids_to_insurance_ids(patient_id_to_insurance_id, payer_id_to_patient_ids)
    crosswalk['payer_id'] = payer_id_to_details
    
    # Validate and correct payer IDs in the crosswalk
    validate_and_correct_payer_ids(client, crosswalk, config)
    
    # Save the crosswalk and log the result
    if save_crosswalk(client, config, crosswalk):
        message = "Crosswalk initialized with mappings for {} payers.".format(len(crosswalk.get('payer_id', {})))
        print(message)
        MediLink_ConfigLoader.log(message, config, level="INFO")
    else:
        print("Failed to save the crosswalk.")
        sys.exit(1)
    
    return crosswalk['payer_id']

def load_and_parse_z_data(config):
    """
    Loads and parses Z data for patient to insurance name mappings from the specified directory.
    
    Args:
        config (dict): Configuration settings for logging.
    
    Returns:
        dict: A mapping of patient IDs to insurance names.
    """
    patient_id_to_insurance_name = {}
    try:
        z_dat_path = config['MediLink_Config']['Z_DAT_PATH']
        MediLink_ConfigLoader.log("Z_DAT_PATH is set to: {}".format(z_dat_path), config, level="DEBUG")
        
        # Get the directory of the Z_DAT_PATH
        directory = os.path.dirname(z_dat_path)
        MediLink_ConfigLoader.log("Looking for .DAT files in directory: {}".format(directory), config, level="DEBUG")
        
        # List all .DAT files in the directory, case insensitive
        dat_files = [f for f in os.listdir(directory) if f.lower().endswith('.dat')]
        MediLink_ConfigLoader.log("Found {} .DAT files in the directory.".format(len(dat_files)), config, level="DEBUG")
        
        # Load processed files tracking
        processed_files_path = os.path.join(directory, 'processed_files.txt')
        if os.path.exists(processed_files_path):
            with open(processed_files_path, 'r') as f:
                processed_files = set(line.strip() for line in f)
            MediLink_ConfigLoader.log("Loaded processed files: {}.".format(processed_files), config, level="DEBUG")
        else:
            processed_files = set()
            MediLink_ConfigLoader.log("No processed files found, starting fresh.", config, level="DEBUG")

        # Filter for new .DAT files that haven't been processed yet, but always include Z.DAT and ZM.DAT
        new_dat_files = [f for f in dat_files if f not in processed_files or f.lower() in ['z.dat', 'zm.dat']]
        MediLink_ConfigLoader.log("Identified {} new .DAT files to process.".format(len(new_dat_files)), config, level="INFO")

        for dat_file in new_dat_files:
            file_path = os.path.join(directory, dat_file)
            MediLink_ConfigLoader.log("Parsing .DAT file: {}".format(file_path), config, level="DEBUG")
            # Parse each .DAT file and accumulate results
            insurance_name_mapping = MediBot_Preprocessor_lib.parse_z_dat(file_path, config['MediLink_Config'])
            if insurance_name_mapping:  # Ensure insurance_name_mapping is not empty
                patient_id_to_insurance_name.update(insurance_name_mapping)

            # Mark this file as processed
            with open(processed_files_path, 'a') as f:
                f.write(dat_file + '\n')
            MediLink_ConfigLoader.log("Marked file as processed: {}".format(dat_file), config, level="DEBUG")

        if not patient_id_to_insurance_name:  # Check if the result is empty
            raise ValueError("Parsed Z data is empty, possibly indicating an error in parsing or all files already processed.") # TODO Add differentiator here because this is dumb.
        MediLink_ConfigLoader.log("Successfully parsed Z data with {} mappings found.".format(len(patient_id_to_insurance_name)), config, level="INFO")
        return patient_id_to_insurance_name  # Ensure the function returns the mapping
    except Exception as e:
        MediLink_ConfigLoader.log("Error loading and parsing Z data: {}".format(e), config, level="ERROR")
        return {}

def crosswalk_update(client, config, crosswalk, skip_known_payers=True): # Upstream of this is only MediBot_Preprocessor.py and MediBot.py
    """
    Updates the crosswalk with insurance data and historical mappings. 
    It loads insurance data, historical payer mappings, and updates the crosswalk accordingly.
    
    Args:
        config (dict): Configuration settings for logging.
        crosswalk (dict): The crosswalk dictionary to update.
        skip_known_payers (bool): If True, skips records with 'name' not equal to 'Unknown'.
    
    Returns:
        bool: True if the crosswalk was updated successfully, False otherwise.
    """
    MediLink_ConfigLoader.log("Starting crosswalk update process...", config, level="INFO")

    # Load insurance data from MAINS
    try:
        MediLink_ConfigLoader.log("Attempting to load insurance data from MAINS...", config, level="DEBUG")
        insurance_name_to_id = MediBot_Preprocessor_lib.load_insurance_data_from_mains(config)
        MediLink_ConfigLoader.log("Loaded insurance data from MAINS with {} entries.".format(len(insurance_name_to_id)), config, level="INFO")
    except Exception as e:
        MediLink_ConfigLoader.log("Error loading insurance data from MAINS: {}".format(e), config, level="ERROR")
        return False

    # Load historical payer to patient mappings
    try:
        MediLink_ConfigLoader.log("Attempting to load historical payer to patient mappings...", config, level="DEBUG")
        patient_id_to_payer_id = MediBot_Preprocessor_lib.load_historical_payer_to_patient_mappings(config)
        MediLink_ConfigLoader.log("Loaded historical mappings with {} entries.".format(len(patient_id_to_payer_id)), config, level="INFO")
    except Exception as e:
        MediLink_ConfigLoader.log("Error loading historical mappings: {}".format(e), config, level="ERROR")
        return False

    # Parse Z data for patient to insurance name mappings
    try:
        patient_id_to_insurance_name = load_and_parse_z_data(config)
        mapping_count = len(patient_id_to_insurance_name) if patient_id_to_insurance_name is not None else 0
        MediLink_ConfigLoader.log("Parsed Z data with {} mappings found.".format(mapping_count), config, level="INFO")
    except Exception as e:
        MediLink_ConfigLoader.log("Error parsing Z data in crosswalk update: {}".format(e), config, level="ERROR")
        return False

    # Check if 'payer_id' key exists and is not empty
    MediLink_ConfigLoader.log("Checking for 'payer_id' key in crosswalk...", config, level="DEBUG")
    if 'payer_id' not in crosswalk or not crosswalk['payer_id']:
        MediLink_ConfigLoader.log("The 'payer_id' list is empty or missing.", config, level="WARNING")
        user_input = input(
            "The 'payer_id' list is empty or missing. Would you like to initialize the crosswalk? (yes/no): "
        ).strip().lower()
        if user_input in ['yes', 'y']:
            MediLink_ConfigLoader.log("User chose to initialize the crosswalk.", config, level="INFO")
            initialize_crosswalk_from_mapat(client, config, crosswalk)
            return True  # Indicate that the crosswalk was initialized
        else:
            MediLink_ConfigLoader.log("User opted not to initialize the crosswalk.", config, level="WARNING")
            return False  # Indicate that the update was not completed

    # Update the crosswalk with new payer IDs and insurance IDs
    for patient_id, payer_id in patient_id_to_payer_id.items():
        """ TODO this needs to be implemented at some point so we can skip known entities.
        # Skip known payers if the flag is set
        if skip_known_payers:
            payer_id_str = next(iter(payer_id))  # Extract the single payer_id from the set
            MediLink_ConfigLoader.log("Checking if payer_id '{}' is known...".format(payer_id_str), config, level="DEBUG")
            payer_info = crosswalk['payer_id'].get(payer_id_str, {})
            payer_name = payer_info.get('name', "Unknown")
            MediLink_ConfigLoader.log("Retrieved payer name: '{}' for payer_id '{}'.".format(payer_name, payer_id), config, level="DEBUG")
            
            if payer_name != "Unknown":
                MediLink_ConfigLoader.log("Skipping known payer_id: '{}' as it is already in the crosswalk.".format(payer_id), config, level="DEBUG")
                continue  # Skip this payer_id
            MediLink_ConfigLoader.log("Skipping known payer_id: {} as it is already in the crosswalk.".format(payer_id), config, level="DEBUG")
            continue  # Skip this payer_id
        """
   
        insurance_name = patient_id_to_insurance_name.get(patient_id)
        if insurance_name and insurance_name in insurance_name_to_id:
            insurance_id = insurance_name_to_id[insurance_name]
           
            # Log the assembly of data
            MediLink_ConfigLoader.log("Assembling data for patient_id '{}': payer_id '{}', insurance_name '{}', insurance_id '{}'.".format(
                patient_id, payer_id, insurance_name, insurance_id), config, level="INFO")
            # Ensure the 'payer_id' key exists in the crosswalk
            if 'payer_id' not in crosswalk:
                crosswalk['payer_id'] = {}
                MediLink_ConfigLoader.log("Initialized 'payer_id' in crosswalk.", config, level="DEBUG")

            # Initialize the payer ID entry if it doesn't exist
            if payer_id not in crosswalk['payer_id']:
                # Prompt the user to select an endpoint name or use the default
                endpoint_options = list(config['MediLink_Config']['endpoints'].keys())
                print("Available endpoints:")
                for idx, key in enumerate(endpoint_options):
                    print("{0}: {1}".format(idx + 1, config['MediLink_Config']['endpoints'][key]['name']))
                user_choice = input("Select an endpoint by number (or press Enter to use the default): ").strip()

                if user_choice.isdigit() and 1 <= int(user_choice) <= len(endpoint_options):
                    selected_endpoint = config['MediLink_Config']['endpoints'][endpoint_options[int(user_choice) - 1]]['name']
                else:
                    selected_endpoint = config['MediLink_Config']['endpoints'][endpoint_options[0]]['name']
                    MediLink_ConfigLoader.log("User opted for default endpoint: {}".format(selected_endpoint), config, level="INFO")

                crosswalk['payer_id'][payer_id] = {
                    'endpoint': selected_endpoint,
                    'medisoft_id': set(),  # Use set to store multiple IDs
                    'medisoft_medicare_id': set()
                }
                MediLink_ConfigLoader.log("Initialized payer ID {} in crosswalk with endpoint '{}'.".format(payer_id, selected_endpoint), config, level="DEBUG")

            # Add the insurance ID to the payer ID entry
            crosswalk['payer_id'][payer_id]['medisoft_id'].add(str(insurance_id))  # Ensure IDs are strings
            MediLink_ConfigLoader.log(
                "Added new insurance ID {} to payer ID {}.".format(insurance_id, payer_id),
                config,
                level="INFO"
            )

            # Log the update of the crosswalk
            MediLink_ConfigLoader.log("Updated crosswalk for payer_id '{}': added insurance_id '{}'.".format(payer_id, insurance_id), config, level="DEBUG")

            # Fetch and store the payer name
            MediLink_ConfigLoader.log("Fetching and storing payer name for payer_id: {}".format(payer_id), config, level="DEBUG")
            fetch_and_store_payer_name(client, payer_id, crosswalk, config)
            MediLink_ConfigLoader.log("Successfully fetched and stored payer name for payer_id: {}".format(payer_id), config, level="INFO")

    # Validate and correct payer IDs in the crosswalk
    MediLink_ConfigLoader.log("Validating and correcting payer IDs in the crosswalk.", config, level="DEBUG")
    validate_and_correct_payer_ids(client, crosswalk, config)

    # Check for any entries marked as "Unknown" and validate them
    unknown_payers = [
        payer_id for payer_id, details in crosswalk.get('payer_id', {}).items()
        if details.get('name') == "Unknown"
    ]
    MediLink_ConfigLoader.log("Found {} unknown payer(s) to validate.".format(len(unknown_payers)), config, level="INFO")
    for payer_id in unknown_payers:
        MediLink_ConfigLoader.log("Fetching and storing payer name for unknown payer_id: {}".format(payer_id), config, level="DEBUG")
        fetch_and_store_payer_name(client, payer_id, crosswalk, config)
        MediLink_ConfigLoader.log("Successfully fetched and stored payer name for unknown payer_id: {}".format(payer_id), config, level="INFO")

    # Ensure multiple medisoft_id values are preserved by converting sets to sorted lists
    for payer_id, details in crosswalk.get('payer_id', {}).items():
        if isinstance(details.get('medisoft_id'), set):
            crosswalk['payer_id'][payer_id]['medisoft_id'] = sorted(list(details['medisoft_id']))
            MediLink_ConfigLoader.log("Converted medisoft_id for payer ID {} to sorted list.".format(payer_id), config, level="DEBUG")
        if isinstance(details.get('medisoft_medicare_id'), set):
            crosswalk['payer_id'][payer_id]['medisoft_medicare_id'] = sorted(list(details['medisoft_medicare_id']))
            MediLink_ConfigLoader.log("Converted medisoft_medicare_id for payer ID {} to sorted list.".format(payer_id), config, level="DEBUG")

    MediLink_ConfigLoader.log("Crosswalk update process completed. Processed {} payer IDs.".format(len(patient_id_to_payer_id)), config, level="INFO")
    return save_crosswalk(client, config, crosswalk)

def update_crosswalk_with_corrected_payer_id(client, old_payer_id, corrected_payer_id, config=None, crosswalk=None): 
    """
    Updates the crosswalk by replacing an old payer ID with a corrected payer ID.
    
    Args:
        old_payer_id (str): The old payer ID to be replaced.
        corrected_payer_id (str): The new payer ID to replace the old one.
        config (dict, optional): Configuration settings for logging.
        crosswalk (dict, optional): The crosswalk dictionary to update.
    
    Returns:
        bool: True if the crosswalk was updated successfully, False otherwise.
    """
    # Ensure full configuration and crosswalk are loaded
    config, crosswalk = ensure_full_config_loaded(config, crosswalk)
        
    # Convert to a regular dict if crosswalk['payer_id'] is an OrderedDict
    if isinstance(crosswalk['payer_id'], dict) and hasattr(crosswalk['payer_id'], 'items'):
        crosswalk['payer_id'] = dict(crosswalk['payer_id'])
    
    MediLink_ConfigLoader.log("Checking if old Payer ID {} exists in crosswalk.".format(old_payer_id), config, level="DEBUG")
    
    MediLink_ConfigLoader.log("Attempting to replace old Payer ID {} with corrected Payer ID {}.".format(old_payer_id, corrected_payer_id), config, level="DEBUG")
        
    # Check if the old payer ID exists before attempting to replace
    if old_payer_id in crosswalk['payer_id']:
        MediLink_ConfigLoader.log("Old Payer ID {} found. Proceeding with replacement.".format(old_payer_id), config, level="DEBUG")
            
        # Store the details of the old payer ID
        old_payer_details = crosswalk['payer_id'][old_payer_id]
        MediLink_ConfigLoader.log("Storing details of old Payer ID {}: {}".format(old_payer_id, old_payer_details), config, level="DEBUG")
        
        # Replace the old payer ID with the corrected one
        crosswalk['payer_id'][corrected_payer_id] = old_payer_details
        MediLink_ConfigLoader.log("Replaced old Payer ID {} with corrected Payer ID {}.".format(old_payer_id, corrected_payer_id), config, level="INFO")
        
        # Remove the old payer ID from the crosswalk
        del crosswalk['payer_id'][old_payer_id]
        MediLink_ConfigLoader.log("Removed old Payer ID {} from crosswalk.".format(old_payer_id), config, level="DEBUG")
    
        # Fetch and store the payer name for the corrected ID
        if fetch_and_store_payer_name(client, corrected_payer_id, crosswalk, config):
            MediLink_ConfigLoader.log("Successfully fetched and stored payer name for corrected Payer ID {}.".format(corrected_payer_id), config, level="INFO")
        else:
            MediLink_ConfigLoader.log("Corrected Payer ID {} updated without a valid name.".format(corrected_payer_id), config, level="WARNING")
        
        # Update csv_replacements
        crosswalk.setdefault('csv_replacements', {})[old_payer_id] = corrected_payer_id
        MediLink_ConfigLoader.log("Updated csv_replacements: {} -> {}.".format(old_payer_id, corrected_payer_id), config, level="INFO")
        print("csv_replacements updated: '{}' -> '{}'.".format(old_payer_id, corrected_payer_id))
        
        return save_crosswalk(client, config, crosswalk)
    else:
        MediLink_ConfigLoader.log("Failed to update crosswalk: old Payer ID {} not found.".format(old_payer_id), config, level="ERROR")
        print("Failed to update crosswalk: could not find old Payer ID '{}'.".format(old_payer_id))
        return False

def update_crosswalk_with_new_payer_id(client, insurance_name, payer_id, config, crosswalk): 
    """
    Updates the crosswalk with a new payer ID for a given insurance name.
    
    Args:
        insurance_name (str): The name of the insurance to associate with the new payer ID.
        payer_id (str): The new payer ID to be added.
        config (dict): Configuration settings for logging.
    """
    # Ensure full configuration and crosswalk are loaded
    config, crosswalk = ensure_full_config_loaded(config, crosswalk)
    
    try:
        # Check if 'payer_id' is present in the crosswalk
        if 'payer_id' not in crosswalk or not crosswalk['payer_id']:
            # Reload the crosswalk if 'payer_id' is missing or empty
            _, crosswalk = MediLink_ConfigLoader.load_configuration(None, config.get('crosswalkPath', 'crosswalk.json'))
            MediLink_ConfigLoader.log("Reloaded crosswalk configuration from {}.".format(config.get('crosswalkPath', 'crosswalk.json')), config, level="DEBUG")
    except KeyError as e:  # Handle KeyError for crosswalk
        MediLink_ConfigLoader.log("KeyError while checking or reloading crosswalk: {}".format(e), config, level="ERROR")
        print("KeyError while checking or reloading crosswalk in update_crosswalk_with_new_payer_id: {}".format(e))
        return False
    except Exception as e:
        MediLink_ConfigLoader.log("Error while checking or reloading crosswalk: {}".format(e), config, level="ERROR")
        print("Error while checking or reloading crosswalk in update_crosswalk_with_new_payer_id: {}".format(e))
        return False
    
    # Load the Medisoft ID for the given insurance name
    try:
        medisoft_id = MediBot_Preprocessor_lib.load_insurance_data_from_mains(config).get(insurance_name)
    except KeyError as e:  # Handle KeyError for config
        MediLink_ConfigLoader.log("KeyError while loading Medisoft ID: {}".format(e), config, level="ERROR")
        print("KeyError while loading Medisoft ID for insurance name {}: {}".format(insurance_name, e))
        return False

    MediLink_ConfigLoader.log("Retrieved Medisoft ID for insurance name {}: {}.".format(insurance_name, medisoft_id), config, level="DEBUG")
    # print("DEBUG: Retrieved Medisoft ID for insurance name {}: {}.".format(insurance_name, medisoft_id))
    
    if medisoft_id:
        medisoft_id_str = str(medisoft_id)
        MediLink_ConfigLoader.log("Processing to update crosswalk with new payer ID: {} for insurance name: {}.".format(payer_id, insurance_name), config, level="DEBUG")
        
        # Initialize the payer ID entry if it doesn't exist
        if payer_id not in crosswalk['payer_id']:
            selected_endpoint = select_endpoint(config)  # Use the helper function to select the endpoint

            # Ensure the 'payer_id' key exists in the crosswalk
            crosswalk['payer_id'][payer_id] = {
                'endpoint': selected_endpoint,
                'medisoft_id': set(),
                'medisoft_medicare_id': []
            }
            MediLink_ConfigLoader.log("Initialized payer ID {} in crosswalk with endpoint '{}'.".format(payer_id, selected_endpoint), config, level="DEBUG")
        else:
            # Check if the existing endpoint is valid
            current_endpoint = crosswalk['payer_id'][payer_id].get('endpoint', None)
            if current_endpoint and current_endpoint not in config['MediLink_Config']['endpoints']:
                print("WARNING: The current endpoint '{}' for payer ID '{}' is not valid.".format(current_endpoint, payer_id))
                MediLink_ConfigLoader.log("Current endpoint '{}' for payer ID '{}' is not valid. Prompting for selection.".format(current_endpoint, payer_id), config, level="WARNING")
                selected_endpoint = select_endpoint(config, current_endpoint)  # Prompt user to select a valid endpoint
                crosswalk['payer_id'][payer_id]['endpoint'] = selected_endpoint  # Update the endpoint in the crosswalk
                MediLink_ConfigLoader.log("Updated payer ID {} with new endpoint '{}'.".format(payer_id, selected_endpoint), config, level="INFO")
            else:
                selected_endpoint = current_endpoint  # Use the existing valid endpoint

        # Add the insurance ID to the payer ID entry - with error handling for the .add() operation
        try:
            if not isinstance(crosswalk['payer_id'][payer_id]['medisoft_id'], set):
                # Convert to set if it's not already one
                crosswalk['payer_id'][payer_id]['medisoft_id'] = set(crosswalk['payer_id'][payer_id]['medisoft_id'])
                MediLink_ConfigLoader.log("Converted medisoft_id to set for payer ID {}.".format(payer_id), config, level="DEBUG")
            
            crosswalk['payer_id'][payer_id]['medisoft_id'].add(str(medisoft_id_str)) # Ensure IDs are strings
            MediLink_ConfigLoader.log(
                "Added new insurance ID {} to payer ID {}.".format(medisoft_id_str, payer_id),
                config,
                level="INFO"
            )
        except AttributeError as e:
            MediLink_ConfigLoader.log("AttributeError while adding medisoft_id: {}".format(e), config, level="ERROR")
            print("Error adding medisoft_id for payer ID {}: {}".format(payer_id, e))
            return False
        
        # Fetch and store the payer name for the new payer ID
        if fetch_and_store_payer_name(client, payer_id, crosswalk, config):
            MediLink_ConfigLoader.log("Successfully fetched and stored payer name for new payer ID {}.".format(payer_id), config, level="INFO")
            MediLink_ConfigLoader.log("Updated crosswalk with new payer ID {} for insurance name {}.".format(payer_id, insurance_name), config, level="INFO")
        else:
            MediLink_ConfigLoader.log("Added new payer ID {} without a valid name for insurance name {}.".format(payer_id, insurance_name), config, level="WARNING")
        
        # Save the updated crosswalk
        save_crosswalk(client, config, crosswalk)
        MediLink_ConfigLoader.log("Crosswalk saved successfully after updating payer ID {}.".format(payer_id), config, level="DEBUG")
    else:
        message = "Failed to update crosswalk: Medisoft ID not found for insurance name {}.".format(insurance_name)
        print(message)
        MediLink_ConfigLoader.log(message, config, level="ERROR")

def save_crosswalk(client, config, crosswalk):
    """
    Saves the crosswalk to a JSON file. Ensures that all necessary keys are present and logs the outcome.
    
    Args:
        config (dict): Configuration settings for logging.
        crosswalk (dict): The crosswalk dictionary to save.
    
    Returns:
        bool: True if the crosswalk was saved successfully, False otherwise.
    """
    try:
        # Determine the path to save the crosswalk
        crosswalk_path = config['MediLink_Config']['crosswalkPath']
        MediLink_ConfigLoader.log("Determined crosswalk path: {}.".format(crosswalk_path), config, level="DEBUG")
    except KeyError:
        crosswalk_path = config.get('crosswalkPath', 'crosswalk.json')
        MediLink_ConfigLoader.log("Using default crosswalk path: {}.".format(crosswalk_path), config, level="DEBUG")
    
    # Validate endpoints for each payer ID in the crosswalk
    for payer_id, details in crosswalk.get('payer_id', {}).items():
        current_endpoint = details.get('endpoint', None)
        if current_endpoint and current_endpoint not in config['MediLink_Config']['endpoints']:
            print("WARNING: The current endpoint '{}' for payer ID '{}' is not valid.".format(current_endpoint, payer_id))
            MediLink_ConfigLoader.log("Current endpoint '{}' for payer ID '{}' is not valid. Prompting for selection.".format(current_endpoint, payer_id), config, level="WARNING")
            selected_endpoint = select_endpoint(config, current_endpoint)  # Prompt user to select a valid endpoint
            crosswalk['payer_id'][payer_id]['endpoint'] = selected_endpoint  # Update the endpoint in the crosswalk
            MediLink_ConfigLoader.log("Updated payer ID {} with new endpoint '{}'.".format(payer_id, selected_endpoint), config, level="INFO")
    
    try:
        # Initialize the 'payer_id' key if it doesn't exist
        if 'payer_id' not in crosswalk: 
            print("save_crosswalk is initializing 'payer_id' key...")
            crosswalk['payer_id'] = {}
            MediLink_ConfigLoader.log("Initialized 'payer_id' key in crosswalk.", config, level="INFO")
        
        # Ensure all payer IDs have a name and initialize medisoft_id and medisoft_medicare_id as empty lists if they do not exist
        for payer_id in crosswalk['payer_id']:
            if 'name' not in crosswalk['payer_id'][payer_id]: 
                fetch_and_store_payer_name(client, payer_id, crosswalk, config)
                MediLink_ConfigLoader.log("Fetched and stored payer name for payer ID: {}.".format(payer_id), config, level="DEBUG")
            
            # Check for the endpoint key
            if 'endpoint' not in crosswalk['payer_id'][payer_id]:
                crosswalk['payer_id'][payer_id]['endpoint'] = select_endpoint(config)  # Use the helper function to set the endpoint
                MediLink_ConfigLoader.log("Initialized 'endpoint' for payer ID {}.".format(payer_id), config, level="DEBUG")

            # Initialize medisoft_id and medisoft_medicare_id as empty lists if they do not exist
            crosswalk['payer_id'][payer_id].setdefault('medisoft_id', [])
            crosswalk['payer_id'][payer_id].setdefault('medisoft_medicare_id', []) # does this work in 3.4.4?
            MediLink_ConfigLoader.log("Ensured 'medisoft_id' and 'medisoft_medicare_id' for payer ID {} are initialized.".format(payer_id), config, level="DEBUG")
        
        # Convert sets to sorted lists for JSON serialization
        for payer_id, details in crosswalk.get('payer_id', {}).items():
            if isinstance(details.get('medisoft_id'), set): 
                crosswalk['payer_id'][payer_id]['medisoft_id'] = sorted(list(details['medisoft_id']))
                MediLink_ConfigLoader.log("Converted medisoft_id for payer ID {} to sorted list.".format(payer_id), config, level="DEBUG")
            if isinstance(details.get('medisoft_medicare_id'), set): 
                crosswalk['payer_id'][payer_id]['medisoft_medicare_id'] = sorted(list(details['medisoft_medicare_id']))
                MediLink_ConfigLoader.log("Converted medisoft_medicare_id for payer ID {} to sorted list.".format(payer_id), config, level="DEBUG")
        
        # Write the crosswalk to the specified file
        with open(crosswalk_path, 'w') as file:
            json.dump(crosswalk, file, indent=4)
        
        MediLink_ConfigLoader.log(
            "Crosswalk saved successfully to {}.".format(crosswalk_path),
            config,
            level="INFO"
        )
        print("Crosswalk saved successfully to {}.".format(crosswalk_path))
        return True
    except KeyError as e:
        print("Key Error: A required key is missing in the crosswalk data - {}.".format(e))
        MediLink_ConfigLoader.log("Key Error while saving crosswalk: {}.".format(e), config, level="ERROR")
        return False
    except TypeError as e:
        print("Type Error: There was a type issue with the data being saved in the crosswalk - {}.".format(e))
        MediLink_ConfigLoader.log("Type Error while saving crosswalk: {}.".format(e), config, level="ERROR")
        return False
    except IOError as e:
        print("I/O Error: An error occurred while writing to the crosswalk file - {}.".format(e))
        MediLink_ConfigLoader.log("I/O Error while saving crosswalk: {}.".format(e), config, level="ERROR")
        return False
    except Exception as e:
        print("Unexpected crosswalk error: {}.".format(e))
        MediLink_ConfigLoader.log("Unexpected error while saving crosswalk: {}.".format(e), config, level="ERROR")
        return False

def select_endpoint(config, current_endpoint=None):
    # BUG Check upstream for the config. One of these is not being passed correctly so we're having to do this check here.
    """
    Prompts the user to select an endpoint from the available options or returns the default endpoint.
    Validates the current endpoint against the available options.

    Args:
        config (dict): Configuration settings for logging. Can be either the full config or config['MediLink_Config'].
        current_endpoint (str, optional): The current endpoint to validate.

    Returns:
        str: The selected endpoint key.

    Raises:
        ValueError: If the config does not contain valid endpoint information.
    """
    # Determine the effective MediLink_Config
    if 'MediLink_Config' in config:
        medi_link_config = config['MediLink_Config']
        MediLink_ConfigLoader.log("Using 'MediLink_Config' from the provided configuration.", config, level="DEBUG")
    else:
        medi_link_config = config
        MediLink_ConfigLoader.log("Using the provided configuration directly as 'MediLink_Config'.", config, level="DEBUG")

    # Attempt to retrieve endpoint options
    try:
        endpoint_options = list(medi_link_config['endpoints'].keys())
        MediLink_ConfigLoader.log("Successfully retrieved endpoint options.", config, level="DEBUG")
    except KeyError:
        MediLink_ConfigLoader.log("Failed to retrieve endpoint options due to KeyError.", config, level="ERROR")
        raise ValueError("Invalid configuration: 'endpoints' not found in config.")
        

    # Ensure there are available endpoints
    if not endpoint_options:
        MediLink_ConfigLoader.log("No endpoints available in the configuration.", config, level="ERROR")
        raise ValueError("No endpoints available in the configuration.")
    else:
        MediLink_ConfigLoader.log("Available endpoints found in the configuration.", config, level="DEBUG")

    print("Available endpoints:")
    for idx, key in enumerate(endpoint_options):
        # Safely retrieve the endpoint name
        endpoint_name = medi_link_config['endpoints'].get(key, {}).get('name', key)
        print("{0}: {1}".format(idx + 1, endpoint_name))

    # Validate the current endpoint if provided
    if current_endpoint and current_endpoint not in endpoint_options:
        print("WARNING: The current endpoint '{}' is not valid.".format(current_endpoint))
        MediLink_ConfigLoader.log("Current endpoint '{}' is not valid. Prompting for selection.".format(current_endpoint), config, level="WARNING")

    user_choice = input("Select an endpoint by number (or press Enter to use the default): ").strip()

    if user_choice.isdigit() and 1 <= int(user_choice) <= len(endpoint_options):
        selected_endpoint = endpoint_options[int(user_choice) - 1]  # Use the key instead of the name
    else:
        selected_endpoint = endpoint_options[0]  # Default to the first key
        MediLink_ConfigLoader.log("User opted for default endpoint: " + selected_endpoint, config, level="INFO")

    return selected_endpoint

def ensure_full_config_loaded(config=None, crosswalk=None):
    """
    Ensures that the full base configuration and crosswalk are loaded.
    If the base config is not valid or the crosswalk is None, reloads them.

    Args:
        config (dict, optional): The current configuration.
        crosswalk (dict, optional): The current crosswalk.

    Returns:
        tuple: The loaded base configuration and crosswalk.
    """
    MediLink_ConfigLoader.log("Ensuring full configuration and crosswalk are loaded.", level="DEBUG")

    # Reload configuration if necessary
    if config is None or 'MediLink_Config' not in config:
        MediLink_ConfigLoader.log("Base config is missing or invalid. Reloading configuration.", level="WARNING")
        config, crosswalk = MediLink_ConfigLoader.load_configuration()
        MediLink_ConfigLoader.log("Base configuration and crosswalk reloaded.", level="INFO")
    else:
        MediLink_ConfigLoader.log("Base config was correctly passed.", level="DEBUG")

    # Reload crosswalk if necessary
    if crosswalk is None:
        MediLink_ConfigLoader.log("Crosswalk is None. Reloading crosswalk.", level="WARNING")
        _, crosswalk = MediLink_ConfigLoader.load_configuration()  # Reloading to get the crosswalk
        MediLink_ConfigLoader.log("Crosswalk reloaded.", level="INFO")

    return config, crosswalk


# MediLink_UI.py
from datetime import datetime
import os, sys

project_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if project_dir not in sys.path:
    sys.path.append(project_dir)

try:
    import MediLink_ConfigLoader
except ImportError:
    from MediLink import MediLink_ConfigLoader

def display_welcome():
    print("\n" + "-" * 60)
    print("          *~^~*:    Welcome to MediLink!    :*~^~*")
    print("-" * 60)

def display_menu(options):
    print("Menu Options:")
    for i, option in enumerate(options):
        print("{0}. {1}".format(i+1, option))

def get_user_choice():
    return input("Enter your choice: ").strip()

def display_exit_message():
    print("\nExiting MediLink.")

def display_invalid_choice():
    print("Invalid choice. Please select a valid option.")

def display_patient_options(detailed_patient_data):
    """
    Displays a list of patients with their current suggested endpoints, prompting for selections to adjust.
    """
    print("\nPlease select the patients to adjust by entering their numbers separated by commas\n(e.g., 1,3,5):")
    # Can disable this extra print for now because the px list would already be on-screen.
    #for i, data in enumerate(detailed_patient_data, start=1):
    #    patient_info = "{0} ({1}) - {2}".format(data['patient_name'], data['patient_id'], data['surgery_date'])
    #    endpoint = data.get('suggested_endpoint', 'N/A')
    #    print("{:<3}. {:<30} Current Endpoint: {}".format(i, patient_info, endpoint))

def get_selected_indices(patient_count):
    """
    Collects user input for selected indices to adjust endpoints.
    """
    selected_indices_input = input("> ")
    selected_indices = [int(index.strip()) - 1 for index in selected_indices_input.split(',') if index.strip().isdigit() and 0 <= int(index.strip()) - 1 < patient_count]
    return selected_indices

def display_patient_for_adjustment(patient_name, suggested_endpoint):
    """
    Displays the current endpoint for a selected patient and prompts for a change.
    """
    print("\n- {0} | Current Endpoint: {1}".format(patient_name, suggested_endpoint))

def get_endpoint_decision():
    """
    Asks the user if they want to change the endpoint.
    Ensures a valid entry of 'Y' or 'N'.
    """
    while True:
        decision = input("Change endpoint? (Y/N): ").strip().lower()
        if decision in ['y', 'n']:
            return decision
        else:
            print("Invalid input. Please enter 'Y' for Yes or 'N' for No.")

def display_endpoint_options(endpoints_config):
    """
    Displays the endpoint options to the user based on the provided mapping.

    Args:
        endpoints_config (dict): A dictionary mapping endpoint keys to their properties, 
                                 where each property includes a 'name' key for the user-friendly name.
                                 Example: {'Availity': {'name': 'Availity'}, 'OptumEDI': {'name': 'OptumEDI'}, ...}

    Returns:
        None
    """
    print("Select the new endpoint for the patient:")
    for index, (key, details) in enumerate(endpoints_config.items(), 1):
        print("{0}. {1}".format(index, details['name']))

def get_new_endpoint_choice():
    """
    Gets the user's choice for a new endpoint.
    Ensures a valid entry of endpoint numbers.
    """
    while True:
        choice = input("Select desired endpoint (e.g. 1, 2): ").strip()
        
        # Auto-correct periods and spaces
        corrected_choice = choice.replace('.', '').replace(' ', '')
        
        if all(index.isdigit() for index in corrected_choice.split(',')):
            return corrected_choice
        else:
            print("Invalid input. I understood your request as: '{}'. Please enter valid endpoint numbers separated by commas.".format(choice))
            print("Tip: Ensure there are no extra spaces or periods in your input.")

# Function to display full list of insurance options
def display_insurance_options(insurance_options=None):

    # Retrieve insurance options with codes and descriptions
    config, _ = MediLink_ConfigLoader.load_configuration()
    insurance_options = config['MediLink_Config'].get('insurance_options')
    
    print("\nInsurance Type Options:")
    for code, description in sorted(insurance_options.items()):
        print("{}: {}".format(code, description))
    print()  # Add a blank line for better readability

def display_patient_summaries(detailed_patient_data):
    """
    Displays summaries of all patients and their suggested endpoints.
    """
    print("\nSummary of patient details and suggested endpoint:")
    for index, summary in enumerate(detailed_patient_data, start=1):
        try:
            display_file_summary(index, summary)
        except KeyError as e:
            print("Summary at index {} is missing key: {}".format(index, e))
    print() # add blank line for improved readability.

def display_file_summary(index, summary):
    # Ensure surgery_date is converted to a datetime object
    surgery_date = datetime.strptime(summary['surgery_date'], "%m-%d-%y")
    
    # Add header row if it's the first index
    if index == 1:
        print("{:<3} {:5} {:<10} {:20} {:15} {:2} {:20}".format(
            "No.", "Date", "ID", "Name", "Primary Ins.", "IT", "Suggested Endpoint"
        ))
        print("-"*80)

    # Check if insurance_type is available; if not, set a default placeholder (this should already be '12' at this point)
    insurance_type = summary.get('insurance_type', '--')

    # Displays the summary of a file.
    print("{:02d}. {:5} ({:<8}) {:20} {:15} {:2} {:20}".format(
        index,
        surgery_date.strftime("%m-%d"),
        summary['patient_id'],
        summary['patient_name'][:20],
        summary['primary_insurance'][:15],
        insurance_type[:2],
        summary['suggested_endpoint'][:20])
    )

def user_select_files(file_list):
    # Sort files by creation time in descending order
    file_list = sorted(file_list, key=os.path.getctime, reverse=True)[:10]  # Limit to max 10 files
    
    print("\nSelect the Z-form files to submit from the following list:\n")

    formatted_files = []
    for i, file in enumerate(file_list):
        basename = os.path.basename(file)
        parts = basename.split('_')
        
        # Try to parse the timestamp from the filename
        if len(parts) > 2:
            try:
                timestamp_str = parts[1] + parts[2].split('.')[0]
                timestamp = datetime.strptime(timestamp_str, '%Y%m%d%H%M%S')
                formatted_date = timestamp.strftime('%m/%d %I:%M %p')  # Changed to 12HR format with AM/PM
            except ValueError:
                formatted_date = basename  # Fallback to original filename if parsing fails
        else:
            formatted_date = basename  # Fallback to original filename if no timestamp
        
        formatted_files.append((formatted_date, file))
        print("{}: {}".format(i + 1, formatted_date))
    
    selected_indices = input("\nEnter the numbers of the files to process, separated by commas\n(or press Enter to select all): ")
    if not selected_indices:
        return [file for _, file in formatted_files]
    
    selected_indices = [int(i.strip()) - 1 for i in selected_indices.split(',')]
    selected_files = [formatted_files[i][1] for i in selected_indices]
    
    return selected_files


#update_medicafe.py
import subprocess, sys, requests, time, pkg_resources, platform

def get_installed_version(package):
    try:
        process = subprocess.Popen(
            [sys.executable, '-m', 'pip', 'show', package],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = process.communicate()
        if process.returncode == 0:
            for line in stdout.decode().splitlines():
                if line.startswith("Version:"):
                    return line.split(":", 1)[1].strip()
        return None
    except Exception as e:
        print("Error retrieving installed version: {}".format(e))
        return None

def get_latest_version(package, retries=3, delay=1):
    """
    Fetch the latest version of the specified package from PyPI with retries.
    """
    for attempt in range(1, retries + 1):
        try:
            response = requests.get("https://pypi.org/pypi/{}/json".format(package), timeout=10)
            response.raise_for_status()  # Raise an error for bad responses
            data = response.json()
            latest_version = data['info']['version']
            
            # Print the version with attempt information
            if attempt == 1:
                print("Latest available version: {}".format(latest_version))
            else:
                print("Latest available version: {} ({} attempt)".format(latest_version, attempt))
            
            # Check if the latest version is different from the current version
            current_version = get_installed_version(package)
            if current_version and compare_versions(latest_version, current_version) == 0:
                # If the versions are the same, perform a second request
                time.sleep(delay)
                response = requests.get("https://pypi.org/pypi/{}/json".format(package), timeout=10)
                response.raise_for_status()
                data = response.json()
                latest_version = data['info']['version']
            
            return latest_version  # Return the version after the check
        except requests.RequestException as e:
            print("Attempt {}: Error fetching latest version: {}".format(attempt, e))
            if attempt < retries:
                print("Retrying in {} seconds...".format(delay))
                time.sleep(delay)
    return None

def check_internet_connection():
    try:
        requests.get("http://www.google.com", timeout=5)
        return True
    except requests.ConnectionError:
        return False

def compare_versions(version1, version2):
    v1_parts = list(map(int, version1.split(".")))
    v2_parts = list(map(int, version2.split(".")))
    return (v1_parts > v2_parts) - (v1_parts < v2_parts)

def upgrade_package(package, retries=3, delay=2):  # Updated retries to 3
    """
    Attempts to upgrade the package multiple times with delays in between.
    """
    if not check_internet_connection():
        print("Error: No internet connection detected. Please check your internet connection and try again.")
        time.sleep(3)  # Pause for 3 seconds before exiting
        sys.exit(1)
    
    for attempt in range(1, retries + 1):
        print("Attempt {} to upgrade {}...".format(attempt, package))
        process = subprocess.Popen(
            [
                sys.executable, '-m', 'pip', 'install', '--upgrade',
                package, '--no-cache-dir', '--disable-pip-version-check', '-q'
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        
        stdout, stderr = process.communicate()
        
        if process.returncode == 0:
            print(stdout.decode().strip())
            new_version = get_installed_version(package)  # Get new version after upgrade
            if compare_versions(new_version, get_latest_version(package)) >= 0:  # Compare versions
                if attempt == 1:
                    print("Upgrade succeeded!")
                else:
                    print("Attempt {}: Upgrade succeeded!".format(attempt))
                time.sleep(delay)
                return True
            else:
                print("Upgrade failed. Current version remains: {}".format(new_version))
                if attempt < retries:
                    print("Retrying in {} seconds...".format(delay))
                    time.sleep(delay)
        else:
            print(stderr.decode().strip())
            print("Attempt {}: Upgrade failed.".format(attempt))
            if attempt < retries:
                print("Retrying in {} seconds...".format(delay))
                time.sleep(delay)
    
    print("Error: All upgrade attempts failed.")
    return False

def ensure_dependencies():
    """Ensure all dependencies listed in setup.py are installed and up-to-date."""
    # Don't try to read requirements.txt as it won't be available after installation
    # Instead, hardcode the same dependencies that are in setup.py
    required_packages = [
        'requests==2.21.0',
        'argparse==1.4.0',
        'tqdm==4.14.0',
        'python-docx==0.8.11',
        'PyYAML==5.2',
        'chardet==3.0.4',
        'msal==1.26.0'
    ]

    # Define problematic packages for Windows XP with Python 3.4
    problematic_packages = ['numpy==1.11.3', 'pandas==0.20.0', 'lxml==4.2.0']
    is_windows_py34 = sys.version_info[:2] == (3, 4) and platform.system() == 'Windows'

    if is_windows_py34:
        print("Detected Windows with Python 3.4")
        print("Please ensure the following packages are installed manually:")
        for pkg in problematic_packages:
            package_name, version = pkg.split('==')
            try:
                installed_version = pkg_resources.get_distribution(package_name).version
                print("{} {} is already installed".format(package_name, installed_version))
                if installed_version != version:
                    print("Note: Installed version ({}) differs from required ({})".format(installed_version, version))
                    print("If you experience issues, consider installing version {} manually".format(version))
            except pkg_resources.DistributionNotFound:
                print("{} is not installed".format(package_name))
                print("Please install {}=={} manually using a pre-compiled wheel".format(package_name, version))
                print("Download from: https://www.lfd.uci.edu/~gohlke/pythonlibs/")
                print("Then run: pip install path\\to\\{}-{}-cp34-cp34m-win32.whl".format(package_name, version))
        print("\nContinuing with other dependencies...")
    else:
        # Add problematic packages to the list for non-Windows XP environments
        required_packages.extend(problematic_packages)

    for pkg in required_packages:
        if '==' in pkg:
            package_name, version = pkg.split('==')  # Extract package name and version
        else:
            package_name = pkg
            version = None  # No specific version required

        # Skip problematic packages on Windows XP Python 3.4
        if is_windows_py34 and any(package_name in p for p in problematic_packages):
            continue

        try:
            installed_version = pkg_resources.get_distribution(package_name).version
            if version and installed_version != version:  # Check if installed version matches required version
                print("Current version of {}: {}".format(package_name, installed_version))
                print("Required version of {}: {}".format(package_name, version))
                time.sleep(2)  # Pause for 2 seconds to allow user to read the output
                if not upgrade_package(package_name):  # Attempt to upgrade/downgrade to the required version
                    print("Warning: Failed to upgrade/downgrade {} to version {}.".format(package_name, version))
                    time.sleep(2)  # Pause for 2 seconds after failure message
            elif version and installed_version == version:  # Check if installed version matches required version
                print("All versions match for {}. No changes needed.".format(package_name))
                time.sleep(1)  # Pause for 2 seconds to allow user to read the output
            elif not version:  # If no specific version is required, check for the latest version
                latest_version = get_latest_version(package_name)
                if latest_version and installed_version != latest_version:
                    print("Current version of {}: {}".format(package_name, installed_version))
                    print("Latest version of {}: {}".format(package_name, latest_version))
                    time.sleep(2)  # Pause for 2 seconds to allow user to read the output
                    if not upgrade_package(package_name):
                        print("Warning: Failed to upgrade {}.".format(package_name))
                        time.sleep(2)  # Pause for 2 seconds after failure message
        except pkg_resources.DistributionNotFound:
            print("Package {} is not installed. Attempting to install...".format(package_name))
            time.sleep(2)  # Pause for 2 seconds before attempting installation
            if not upgrade_package(package_name):
                print("Warning: Failed to install {}.".format(package_name))
                time.sleep(2)  # Pause for 2 seconds after failure message

def main():
    # Ensure internet connection before proceeding
    if not check_internet_connection():
        print("Error: No internet connection. Please check your connection and try again.")
        time.sleep(3)  # Pause for 3 seconds before exiting
        sys.exit(1)

    # Ensure all dependencies are met before proceeding
    response = input("Do you want to check dependencies? (yes/no, default/enter is no): ").strip().lower()
    if response in ['yes', 'y']:
        ensure_dependencies()
    else:
        print("Skipping dependency check.")
        time.sleep(3)  # Pause for 3 seconds before proceeding

    package = "medicafe"
    
    current_version = get_installed_version(package)
    if not current_version:
        print("{} is not installed.".format(package))
        time.sleep(3)  # Pause for 3 seconds before exiting
        sys.exit(1)
    
    latest_version = get_latest_version(package)
    if not latest_version:
        print("Could not retrieve the latest version information.")
        time.sleep(3)  # Pause for 3 seconds before exiting
        sys.exit(1)
    
    print("Current version of {}: {}".format(package, current_version))
    print("Latest version of {}: {}".format(package, latest_version))
    
    if compare_versions(latest_version, current_version) > 0:
        print("A newer version is available. Proceeding with upgrade.")
        if upgrade_package(package):
            # Verify upgrade
            time.sleep(3)
            new_version = get_installed_version(package)
            if compare_versions(new_version, latest_version) >= 0:
                print("Upgrade successful. New version: {}".format(new_version))
            else:
                print("Upgrade failed. Current version remains: {}".format(new_version))
                time.sleep(3)  # Pause for 3 seconds before exiting
                sys.exit(1)
        else:
            time.sleep(3)  # Pause for 3 seconds before exiting
            sys.exit(1)
    else:
        print("You already have the latest version installed.")
        time.sleep(3)  # Pause for 3 seconds before exiting
        sys.exit(1)

if __name__ == "__main__":
    main()



# MediLink_Parser.py

import re

def parse_era_content(content, debug=False):
    extracted_data = []
    normalized_content = content.replace('~\n', '~')
    lines = normalized_content.split('~')

    record = {}
    check_eft, payer_address = None, None
    allowed_amount, write_off, patient_responsibility, adjustment_amount = 0, 0, 0, 0
    is_payer_section = False

    for line in lines:
        segments = line.split('*')

        if segments[0] == 'TRN' and len(segments) > 2:
            check_eft = segments[2]

        if segments[0] == 'N1':
            if segments[1] == 'PR':
                is_payer_section = True
            elif segments[1] == 'PE':
                is_payer_section = False

        if is_payer_section and segments[0] == 'N3' and len(segments) > 1:
            payer_address = segments[1]

        if segments[0] == 'CLP' and len(segments) >= 5:
            if record:
                if adjustment_amount == 0 and (write_off > 0 or patient_responsibility > 0):
                    adjustment_amount = write_off + patient_responsibility

                record.update({
                    'Payer Address': payer_address,
                    'Allowed Amount': allowed_amount,
                    'Write Off': write_off,
                    'Patient Responsibility': patient_responsibility,
                    'Adjustment Amount': adjustment_amount,
                })
                extracted_data.append(record)

                allowed_amount, write_off, patient_responsibility, adjustment_amount = 0, 0, 0, 0

            record = {
                'Check EFT': check_eft,
                'Chart Number': segments[1],
                'Payer Address': payer_address,
                'Amount Paid': segments[4],
                'Charge': segments[3],
            }

        elif segments[0] == 'CAS':
            if segments[1] == 'CO':
                write_off += float(segments[3])
            elif segments[1] == 'PR':
                patient_responsibility += float(segments[3])
            elif segments[1] == 'OA':
                adjustment_amount += float(segments[3])

        elif segments[0] == 'AMT' and segments[1] == 'B6':
            allowed_amount += float(segments[2])

        elif segments[0] == 'DTM' and (segments[1] == '232' or segments[1] == '472'):
            record['Date of Service'] = segments[2]

    if record:
        if adjustment_amount == 0 and (write_off > 0 or patient_responsibility > 0):
            adjustment_amount = write_off + patient_responsibility
        record.update({
            'Allowed Amount': allowed_amount,
            'Write Off': write_off,
            'Patient Responsibility': patient_responsibility,
            'Adjustment Amount': adjustment_amount,
        })
        extracted_data.append(record)

    if debug:
        print("Parsed ERA Content:")
        for data in extracted_data:
            print(data)

    return extracted_data

def parse_277_content(content, debug=False):
    segments = content.split('~')
    records = []
    current_record = {}
    for segment in segments:
        parts = segment.split('*')
        if parts[0] == 'HL':
            if current_record:
                records.append(current_record)
                current_record = {}
        elif parts[0] == 'NM1':
            if parts[1] == 'QC':
                current_record['Patient'] = parts[3] + ' ' + parts[4]
            elif parts[1] == '41':
                current_record['Clearing House'] = parts[3]
            elif parts[1] == 'PR':
                current_record['Payer'] = parts[3]
        elif parts[0] == 'TRN':
            current_record['Claim #'] = parts[2]
        elif parts[0] == 'STC':
            current_record['Status'] = parts[1]
            if len(parts) > 4:
                current_record['Paid'] = parts[4]
        elif parts[0] == 'DTP':
            if parts[1] == '472':
                current_record['Serv.'] = parts[3]
            elif parts[1] == '050':
                current_record['Proc.'] = parts[3]
        elif parts[0] == 'AMT':
            if parts[1] == 'YU':
                current_record['Charged'] = parts[2]

    if current_record:
        records.append(current_record)

    if debug:
        print("Parsed 277 Content:")
        for record in records:
            print(record)

    return records

def parse_277IBR_content(content, debug=False):
    return parse_277_content(content, debug)

def parse_277EBR_content(content, debug=False):
    return parse_277_content(content, debug)

def parse_dpt_content(content, debug=False):
    extracted_data = []
    lines = content.splitlines()
    record = {}
    for line in lines:
        if 'Patient Account Number:' in line:
            if record:
                extracted_data.append(record)
            record = {}
        parts = line.split(':')
        if len(parts) == 2:
            key, value = parts[0].strip(), parts[1].strip()
            record[key] = value
    if record:
        extracted_data.append(record)

    if debug:
        print("Parsed DPT Content:")
        for data in extracted_data:
            print(data)

    return extracted_data

def parse_ebt_content(content, debug=False):
    extracted_data = []  # List to hold all extracted records
    lines = content.splitlines()  # Split the content into individual lines
    record = {}  # Dictionary to hold the current record being processed

    # Regular expression pattern to match key-value pairs in the format "Key: Value"
    key_value_pattern = re.compile(r'([^:]+):\s*(.+?)(?=\s{2,}[^:]+?:|$)')

    for line in lines:
        # Check for the start of a new record based on the presence of 'Patient Name'
        if 'Patient Name:' in line and record:
            ebt_post_processor(record)  # Process the current record before adding it to the list
            extracted_data.append(record)  # Add the completed record to the list
            record = {}  # Reset the record for the next entry

        # Find all key-value pairs in the current line
        matches = key_value_pattern.findall(line)
        for key, value in matches:
            key = key.strip()  # Remove leading/trailing whitespace from the key
            value = value.strip()  # Remove leading/trailing whitespace from the value
            record[key] = value  # Add the key-value pair to the current record

    # Process and add the last record if it exists
    if record:
        ebt_post_processor(record)  # Final processing of the last record
        extracted_data.append(record)  # Add the last record to the list

    # Debug output to show parsed data if debugging is enabled
    if debug:
        print("Parsed EBT Content:")
        for data in extracted_data:
            print(data)

    return extracted_data  # Return the list of extracted records

def ebt_post_processor(record):
    # Process the 'Message Initiator' field to separate it from 'Message Type'
    if 'Message Initiator' in record and 'Message Type:' in record['Message Initiator']:
        parts = record['Message Initiator'].split('Message Type:')  # Split the string into parts
        record['Message Initiator'] = parts[0].strip()  # Clean up the 'Message Initiator'
        record['Message Type'] = parts[1].strip()  # Clean up the 'Message Type'

def parse_ibt_content(content, debug=False):
    extracted_data = []
    lines = content.splitlines()
    record = {}
    for line in lines:
        if 'Submitter Batch ID:' in line:
            if record:
                extracted_data.append(record)
            record = {}
        parts = line.split(':')
        if len(parts) == 2:
            key, value = parts[0].strip(), parts[1].strip()
            record[key] = value
    if record:
        extracted_data.append(record)

    if debug:
        print("Parsed IBT Content:")
        for data in extracted_data:
            print(data)

    return extracted_data


import pandas as pd
import re

def load_csv(file_path):
    # Loads a CSV file and returns a pandas DataFrame
    return pd.read_csv(file_path, header=None)

def segment_data(data):
    # Segments the data into individual patient records
    
    patient_records = []
    current_record = []
    
    for line in data.itertuples(index=False):
        # Convert the tuple to a list to process it easier
        line = list(line)
        # Flatten the list and filter out NaN values
        line = [item for item in line if pd.notna(item)]
        
        if line:  # Make sure there is data in the line
            # Check for the delimiter indicating a new patient record
            if 'PATIENT INFORMATION' in line[0]:
                if current_record:
                    # If there's an existing record, this means we've reached a new one
                    # Save the current record and start a new one
                    patient_records.append(current_record)
                    current_record = []
            # Add the line to the current patient record
            current_record.extend(line)
    
    # Don't forget to add the last record after exiting the loop
    if current_record:
        patient_records.append(current_record)
    
    return patient_records

# Function to extract key-value pairs from a patient record segment
def extract_patient_data(patient_record):
    patient_data = {
        "Name": None,
        "Patient ID": None,
        "Address": None,
        "Home Phone": None,
        "DOB": None,
        "Gender": None
    }
    
    # Function to extract value after a specific key in the patient record
    def extract_value_after_key(lines, key):
        for index, line in enumerate(lines):
            if key in line:
                try:
                    split_line = line.split('\n')
                    if len(split_line) > 1:
                        # Return the value only if it exists after the newline character
                        return split_line[1].strip()                
                except AttributeError:
                    # Handle the case where 'line' is not a string and doesn't have the 'split' method
                    print("Error extracting value after key:", line)
                    pass     
    
    # For each key in patient_data, extract its value from the patient_record
    for key in patient_data.keys():
        patient_data[key] = extract_value_after_key(patient_record, key)
    return patient_data


def parse_insurance_info(patient_record):
    insurance_data = {
        "Primary Insurance": None,
        "Primary Policy Number": None,
        "Primary Group Number": None,
        "Secondary Insurance": None,
        "Secondary Policy Number": None,
        "Secondary Group Number": None
    }

    insurance_section_started = False
    secondary_insurance_detected = False
    group_header_detected = False

    for element in patient_record:
        if 'INSURANCE INFORMATION' in element:
            insurance_section_started = True
            secondary_insurance_detected = False
            continue

        if insurance_section_started:
            split_element = element.split('\n')
            if 'Primary Insurance' in element:
                insurance_data["Primary Insurance"] = element.split('\n')[1].strip() if len(element.split('\n')) > 1 else None
            elif 'Secondary Insurance' in element and len(split_element) > 1 and split_element[1].strip():
                insurance_data["Secondary Insurance"] = element.split('\n')[1].strip() if len(element.split('\n')) > 1 else None
                secondary_insurance_detected = True
            elif 'Policy Number' in element:
                split_element = element.split('\n')
                if len(split_element) > 1:
                    if not insurance_data["Primary Policy Number"]:
                        insurance_data["Primary Policy Number"] = split_element[1].strip()
                    elif secondary_insurance_detected and not insurance_data["Secondary Policy Number"]:
                        insurance_data["Secondary Policy Number"] = split_element[1].strip()
            elif 'Group Number' in element:
                #print("Group Detected: ", element, secondary_insurance_detected)
                group_header_detected = not group_header_detected # toggle between T/F to proxy as first or second position.
                split_element = element.split('\n')
                if len(split_element) > 1:
                    if not insurance_data["Primary Group Number"] and group_header_detected:
                        insurance_data["Primary Group Number"] = split_element[1].strip()
                    elif secondary_insurance_detected and not insurance_data["Secondary Group Number"] and not group_header_detected:
                        insurance_data["Secondary Group Number"] = split_element[1].strip()

    return insurance_data

def structure_data(patient_data_list):
    # Define the column headers based on the sample data provided earlier
    column_headers = [
        "Name", 
        "Patient ID", 
        "Address", 
        "Home Phone", 
        "DOB", 
        "Gender", 
        "Primary Insurance", 
        "Primary Policy Number", 
        "Primary Group Number",
        "Secondary Insurance", 
        "Secondary Policy Number",
        "Secondary Group Number"
    ]

    # Initialize a list to hold structured patient records
    structured_patient_records = []

    # Iterate over each patient record in the list
    for patient_record in patient_data_list:
        # Extract the basic patient data
        patient_data = extract_patient_data(patient_record)
        # Extract the insurance information
        insurance_data = parse_insurance_info(patient_record)
        # Merge the two dictionaries
        full_patient_data = {**patient_data, **insurance_data}
        
        # Add the cleaned and transformed data to the list
        structured_patient_records.append(full_patient_data)

    # Create the DataFrame with the structured patient data
    structured_patient_df = pd.DataFrame(structured_patient_records, columns=column_headers)

    # Return the structured DataFrame
    return structured_patient_df

def validate_data(data_frame):
    # Performing Quality Assurance and Validation checks on the structured data

    # Completeness Check: Check for missing values in critical fields
    missing_values_check = data_frame.isnull().sum()

    # Consistency Check: Ensure data formats are consistent
    date_format_check = data_frame['DOB'].apply(lambda x: bool(re.match(r'\d{4}-\d{2}-\d{2}', x)) if pd.notnull(x) else True)
    phone_format_check = data_frame['Home Phone'].apply(lambda x: bool(re.match(r'\+\d-\d{3}-\d{3}-\d{4}', x)) if pd.notnull(x) else True)

    # Anomaly Detection: This can be complex and domain-specific. As a basic check, we can look for outliers in data like dates.
    dob_anomalies_check = data_frame['DOB'].describe()

    # Compile the results of the checks
    validation_results = {
        "Missing Values Check": missing_values_check,
        "Date Format Consistency": all(date_format_check),
        "Phone Format Consistency": all(phone_format_check),
        "DOB Anomalies Check": dob_anomalies_check
    }
    
    print(validation_results)  # Display validation results
    return data_frame  # Return the validated DataFrame


# Main function to orchestrate the cleaning process
def clean_patient_data(file_path):
    # Load the CSV file
    sxpatient_data = load_csv(file_path)

    # Segment the data
    segmented_patient_records = segment_data(sxpatient_data)

    # Structure the data
    structured_data_frame = structure_data(segmented_patient_records)

    # Validate the data
    validated_data = validate_data(structured_data_frame)
    
    return validated_data

# Path to the CSV file with escaped backslashes
file_path_sxpatient = 'C:\\Users\\danie\\OneDrive\\Desktop\\CSV02012024.CSV'
# Define the file path for the output CSV file
output_file_path = 'G:\\My Drive\\CocoWave\\XP typing bot\\cleaned_FEB01SXcsv_group.csv'

# Call the main function to clean the patient data
cleaned_patient_data = clean_patient_data(file_path_sxpatient)

# Display the first few rows of the cleaned and validated data to verify the output
print(cleaned_patient_data.head())

# Save the processed data to a CSV file
cleaned_patient_data.to_csv(output_file_path, index=False)

print(f"Processed data saved to {output_file_path}")

# Development Roadmap

# Do not delete leading zeros from insurance numbers


import csv, os, re, subprocess, time
from datetime import datetime, timedelta

# Need this for running Medibot and MediLink
try:
    import MediLink_ConfigLoader
    import MediLink_UI
except ImportError:
    from . import MediLink_ConfigLoader
    from . import MediLink_UI

# Helper function to slice and strip values with optional key suffix
def slice_data(data, slices, suffix=''):
    # Convert slices list to a tuple for slicing operation
    return {key + suffix: data[slice(*slices[key])].strip() for key in slices}

# Function to parse fixed-width Medisoft output and extract claim data
def parse_fixed_width_data(personal_info, insurance_info, service_info, service_info_2=None, service_info_3=None, config=None):
    
    # Make sure we have the right config
    if not config:  # Checks if config is None or an empty dictionary
        MediLink_ConfigLoader.log("No config passed to parse_fixed_width_data. Re-loading config...", level="WARNING")
        config, _ = MediLink_ConfigLoader.load_configuration()
    
    config = config.get('MediLink_Config', config) # Safest config call.
    
    # Load slice definitions from config within the MediLink_Config section
    personal_slices = config['fixedWidthSlices']['personal_slices']
    insurance_slices = config['fixedWidthSlices']['insurance_slices']
    service_slices = config['fixedWidthSlices']['service_slices']

    # Parse each segment
    parsed_data = {}
    parsed_data.update(slice_data(personal_info, personal_slices))
    parsed_data.update(slice_data(insurance_info, insurance_slices))
    parsed_data.update(slice_data(service_info, service_slices))
    
    if service_info_2:
        parsed_data.update(slice_data(service_info_2, service_slices, suffix='_2'))
    
    if service_info_3:
        parsed_data.update(slice_data(service_info_3, service_slices, suffix='_3'))
    
    # Replace underscores with spaces in first and last names since this is downstream of MediSoft. 
    if 'FIRST' in parsed_data:
        parsed_data['FIRST'] = parsed_data['FIRST'].replace('_', ' ')
    if 'LAST' in parsed_data:
        parsed_data['LAST'] = parsed_data['LAST'].replace('_', ' ')
    
    MediLink_ConfigLoader.log("Successfully parsed data from segments", config, level="INFO")
    
    return parsed_data

# Function to read fixed-width Medisoft output and extract claim data
def read_fixed_width_data(file_path):
    # Reads the fixed width data from the file and yields each patient's
    # personal, insurance, and service information.
    MediLink_ConfigLoader.log("Starting to read fixed width data...")
    with open(file_path, 'r') as file:
        lines_buffer = []  # Buffer to hold lines for current patient data
        
        def yield_record(buffer):
            personal_info = buffer[0]
            insurance_info = buffer[1]
            service_info = buffer[2]
            service_info_2 = buffer[3] if len(buffer) > 3 else None
            service_info_3 = buffer[4] if len(buffer) > 4 else None
            MediLink_ConfigLoader.log("Successfully read data from file: {}".format(file_path), level="INFO")
            return personal_info, insurance_info, service_info, service_info_2, service_info_3
        
        for line in file:
            stripped_line = line.strip()
            if stripped_line:
                lines_buffer.append(stripped_line)
                if 3 <= len(lines_buffer) <= 5:
                    next_line = file.readline().strip()
                    if not next_line:
                        yield yield_record(lines_buffer)
                        lines_buffer.clear()
            else:
                if len(lines_buffer) >= 3:
                    yield yield_record(lines_buffer)
                    lines_buffer.clear()
                    
        if lines_buffer:  # Yield any remaining buffer if file ends without a blank line
            yield yield_record(lines_buffer)

# TODO (Refactor) Consider consolidating with the other read_fixed_with_data 
def read_general_fixed_width_data(file_path, slices):
    # handle any fixed-width data based on provided slice definitions
    with open(file_path, 'r', encoding='utf-8') as file:
        next(file)  # Skip the header
        for line_number, line in enumerate(file, start=1):
            insurance_name = {key: line[start:end].strip() for key, (start, end) in slices.items()}
            yield insurance_name, line_number

def consolidate_csvs(source_directory, file_prefix="Consolidated", interactive=False):
    """
    Consolidate CSV files in the source directory into a single CSV file.
    
    Parameters:
        source_directory (str): The directory containing the CSV files to consolidate.
        file_prefix (str): The prefix for the consolidated file's name.
        interactive (bool): If True, prompt the user for confirmation before overwriting existing files.
    
    Returns:
        str: The filepath of the consolidated CSV file, or None if no files were consolidated.
    """
    today = datetime.now()
    consolidated_filename = "{}_{}.csv".format(file_prefix, today.strftime("%m%d%y"))
    consolidated_filepath = os.path.join(source_directory, consolidated_filename)

    consolidated_data = []
    header_saved = False
    expected_header = None

    # Check if the file already exists and log the action
    if os.path.exists(consolidated_filepath):
        MediLink_ConfigLoader.log("The file {} already exists. It will be overwritten.".format(consolidated_filename), level="INFO")
        if interactive:
            overwrite = input("The file {} already exists. Do you want to overwrite it? (y/n): ".format(consolidated_filename)).strip().lower()
            if overwrite != 'y':
                MediLink_ConfigLoader.log("User opted not to overwrite the file {}.".format(consolidated_filename), level="INFO")
                return None

    for filename in os.listdir(source_directory):
        filepath = os.path.join(source_directory, filename)
        if not filepath.endswith('.csv') or os.path.isdir(filepath) or filepath == consolidated_filepath:
            continue  # Skip non-CSV files, directories, and the target consolidated file itself

        # Check if the file was created within the last day
        modification_time = datetime.fromtimestamp(os.path.getmtime(filepath))
        if modification_time < today - timedelta(days=1):
            continue  # Skip files not modified in the last day

        try:
            with open(filepath, 'r') as csvfile:
                reader = csv.reader(csvfile)
                header = next(reader)  # Read the header
                if not header_saved:
                    expected_header = header
                    consolidated_data.append(header)
                    header_saved = True
                elif header != expected_header:
                    MediLink_ConfigLoader.log("Header mismatch in file {}. Skipping file.".format(filepath), level="WARNING")
                    continue

                consolidated_data.extend(row for row in reader)
        except StopIteration:
            MediLink_ConfigLoader.log("File {} is empty or contains only header. Skipping file.".format(filepath), level="WARNING")
            continue
        except Exception as e:
            MediLink_ConfigLoader.log("Error processing file {}: {}".format(filepath, e), level="ERROR")
            continue

        os.remove(filepath)
        MediLink_ConfigLoader.log("Deleted source file after consolidation: {}".format(filepath), level="INFO")

    if consolidated_data:
        with open(consolidated_filepath, 'w') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(consolidated_data)
        MediLink_ConfigLoader.log("Consolidated CSVs into {}".format(consolidated_filepath), level="INFO")
        return consolidated_filepath
    else:
        MediLink_ConfigLoader.log("No valid CSV files were found for consolidation.", level="INFO")
        return None

def operate_winscp(operation_type, files, endpoint_config, local_storage_path, config):
    """
    General function to operate WinSCP for uploading or downloading files.
    """
    MediLink_ConfigLoader.log("Starting operate_winscp with operation_type: {}".format(operation_type))
    
    config = ensure_config_loaded(config)
    winscp_path = get_winscp_path(config)
    
    if not os.path.isfile(winscp_path):
        MediLink_ConfigLoader.log("WinSCP.com not found at {}".format(winscp_path), level="ERROR")
        return []

    validate_endpoint_config(endpoint_config)
    winscp_log_path = setup_logging(operation_type, local_storage_path)

    # Validate the local_storage_path and replace it if necessary
    local_storage_path = validate_local_storage_path(local_storage_path, config)

    remote_directory = get_remote_directory(endpoint_config, operation_type)
    command = build_command(winscp_path, winscp_log_path, endpoint_config, remote_directory, operation_type, files, local_storage_path)

    if config.get("TestMode", True):
        MediLink_ConfigLoader.log("Test mode is enabled. Simulating operation.")
        return simulate_operation(operation_type, files, config)
    
    result = execute_winscp_command(command, operation_type, files, local_storage_path)
    MediLink_ConfigLoader.log("[Execute WinSCP Command] Result: {}".format(result), level="DEBUG")
    return result

def validate_local_storage_path(local_storage_path, config):
    """
    Validates the local storage path and replaces it with outputFilePath from config if it contains spaces.
    """
    if ' ' in local_storage_path:
        MediLink_ConfigLoader.log("Local storage path contains spaces, using outputFilePath from config.", level="WARN")
        output_file_path = config.get('outputFilePath', None)
        if not output_file_path:
            raise ValueError("outputFilePath not found in config.")
        return os.path.normpath(output_file_path)
    return os.path.normpath(local_storage_path)

def ensure_config_loaded(config):
    MediLink_ConfigLoader.log("Ensuring configuration is loaded.")
    if not config:
        MediLink_ConfigLoader.log("Warning: No config passed to ensure_config_loaded. Re-loading config...")
        config, _ = MediLink_ConfigLoader.load_configuration()
    
    # Check if config was successfully loaded
    if not config or 'MediLink_Config' not in config:
        MediLink_ConfigLoader.log("Failed to load the MediLink configuration. Config is None or missing 'MediLink_Config'.")
        raise RuntimeError("Failed to load the MediLink configuration. Config is None or missing 'MediLink_Config'.")

    # Check that 'endpoints' key exists within 'MediLink_Config'
    if 'endpoints' not in config['MediLink_Config']:
        MediLink_ConfigLoader.log("The loaded configuration is missing the 'endpoints' section.")
        raise ValueError("The loaded configuration is missing the 'endpoints' section.")

    # Additional checks can be added here to ensure all expected keys and structures are present
    if 'local_storage_path' not in config['MediLink_Config']:
        MediLink_ConfigLoader.log("The loaded configuration is missing the 'local_storage_path' setting.")
        raise ValueError("The loaded configuration is missing the 'local_storage_path' setting.")

    MediLink_ConfigLoader.log("Configuration loaded successfully.")
    return config['MediLink_Config']  # Return the relevant part of the config for simplicity

def get_winscp_path(config):
    MediLink_ConfigLoader.log("Retrieving WinSCP path from provided config.")
    
    def find_winscp_path(cfg):
        if 'winscp_path' in cfg:
            # cfg is already 'MediLink_Config'
            MediLink_ConfigLoader.log("Config provided directly as 'MediLink_Config'.")
            return cfg.get('winscp_path')
        else:
            # cfg is the full configuration, retrieve 'MediLink_Config'
            MediLink_ConfigLoader.log("Config provided as full configuration; accessing 'MediLink_Config'.")
            medi_link_config = cfg.get('MediLink_Config', {})
            return medi_link_config.get('winscp_path')

    # Attempt to find the WinSCP path using the provided config
    winscp_path = find_winscp_path(config)
    
    # If the path is not found, attempt to use default paths
    if not winscp_path:
        error_message = "WinSCP path not found in config. Attempting to use default paths."
        # print(error_message)
        MediLink_ConfigLoader.log(error_message)
        
        # Try the default paths
        default_paths = [
            os.path.join(os.getcwd(), "Installers", "WinSCP-Portable", "WinSCP.com"),
            os.path.join(os.getcwd(), "Necessary Programs", "WinSCP-Portable", "WinSCP.com")
        ]
        
        for path in default_paths:
            if os.path.exists(path):
                found_message = "WinSCP found at {}. Using this path.".format(path)
                # print(found_message)
                MediLink_ConfigLoader.log(found_message)
                return path
        
        # If no valid path is found, attempt to reload the configuration
        reload_message = "WinSCP not found in config or default paths. Reloading the entire configuration."
        # print(reload_message)
        MediLink_ConfigLoader.log(reload_message)
        
        try:
            config, _ = MediLink_ConfigLoader.load_configuration()
            winscp_path = find_winscp_path(config)
            
            if winscp_path:
                success_message = "WinSCP path found after reloading configuration. Using this path."
                # print(success_message)
                MediLink_ConfigLoader.log(success_message)
                return winscp_path
            else:
                raise FileNotFoundError("WinSCP path not found even after reloading configuration.")
        
        except Exception as e:
            error_message = "Failed to reload configuration or find WinSCP path: {}. Exiting script.".format(e)
            print(error_message)
            MediLink_ConfigLoader.log(error_message)
            raise FileNotFoundError(error_message)
    
    return winscp_path

def validate_endpoint_config(endpoint_config):
    MediLink_ConfigLoader.log("Validating endpoint configuration.")
    if not isinstance(endpoint_config, dict):
        MediLink_ConfigLoader.log("Endpoint configuration object is invalid. Expected a dictionary, got: {}".format(type(endpoint_config)))
        raise ValueError("Endpoint configuration object is invalid. Expected a dictionary, got: {}".format(type(endpoint_config)))

def setup_logging(operation_type, local_storage_path):
    MediLink_ConfigLoader.log("Setting up logging for operation type: {}".format(operation_type))
    log_filename = "winscp_upload.log" if operation_type == "upload" else "winscp_download.log"
    return os.path.join(local_storage_path, log_filename)

def get_remote_directory(endpoint_config, operation_type):
    MediLink_ConfigLoader.log("Getting remote directory for operation type: {}".format(operation_type))
    if endpoint_config is None:
        MediLink_ConfigLoader.log("Error: Endpoint configuration is None.")
        raise ValueError("Endpoint configuration is None. Expected a dictionary with configuration details.")

    if not isinstance(endpoint_config, dict):
        MediLink_ConfigLoader.log("Error: Endpoint configuration is invalid. Expected a dictionary, got: {}".format(type(endpoint_config)))
        raise TypeError("Endpoint configuration is invalid. Expected a dictionary, got: {}".format(type(endpoint_config)))

    try:
        if operation_type == "upload":
            return endpoint_config['remote_directory_up']
        elif operation_type == "download":
            return endpoint_config['remote_directory_down']
        else:
            MediLink_ConfigLoader.log("Invalid operation type: {}. Expected 'upload' or 'download'.".format(operation_type))
            raise ValueError("Invalid operation type: {}. Expected 'upload' or 'download'.".format(operation_type))
    except KeyError as e:
        MediLink_ConfigLoader.log("Critical Error: Endpoint config is missing key: {}".format(e))
        raise RuntimeError("Configuration error: Missing required remote directory in endpoint configuration.")

def build_command(winscp_path, winscp_log_path, endpoint_config, remote_directory, operation_type, files, local_storage_path, newer_than=None, filemask=None):
    # Log the operation type
    MediLink_ConfigLoader.log("[Build Command] Building WinSCP command for operation type: {}".format(operation_type))

    session_name = endpoint_config.get('session_name', '')

    # Initial command structure with options to disable timestamp preservation and permission setting (should now be compatible with Availity, hopefully this doesn't break everything else)
    command = [
        winscp_path,
        '/log=' + winscp_log_path,
        '/loglevel=1',
        '/nopreservetime',  # Disable timestamp preservation
        '/nopermissions',   # Disable permission setting
        '/command',
        'open {}'.format(session_name),
        'cd /',
        'cd {}'.format(remote_directory)
    ]

    try:
        # Handle upload operation
        if operation_type == "upload":
            if not files:
                MediLink_ConfigLoader.log("Error: No files provided for upload operation.", level="ERROR")
                raise ValueError("No files provided for upload operation.")

            put_commands = []
            for f in files:
                # Normalize the path
                normalized_path = os.path.normpath(f)
                original_path = normalized_path  # Keep for logging

                # Remove leading slash if present
                if normalized_path.startswith('\\') or normalized_path.startswith('/'):
                    normalized_path = normalized_path.lstrip('\\/')
                    MediLink_ConfigLoader.log("Removed leading slash from path: {}".format(original_path), level="DEBUG")

                # Remove trailing slash if present
                if normalized_path.endswith('\\') or normalized_path.endswith('/'):
                    normalized_path = normalized_path.rstrip('\\/')
                    MediLink_ConfigLoader.log("Removed trailing slash from path: {}".format(original_path), level="DEBUG")

                # Determine if quotes are necessary (e.g., if path contains spaces)
                if ' ' in normalized_path:
                    put_command = 'put "{}"'.format(normalized_path)
                    MediLink_ConfigLoader.log("Constructed put command with quotes: {}".format(put_command), level="DEBUG")
                else:
                    put_command = 'put {}'.format(normalized_path)
                    MediLink_ConfigLoader.log("Constructed put command without quotes: {}".format(put_command), level="DEBUG")

                put_commands.append(put_command)
            command += put_commands

        # Handle download operation
        elif operation_type == "download":
            lcd_path = os.path.normpath(local_storage_path)
            original_lcd_path = lcd_path  # Keep for logging

            # Remove leading slash if present
            if lcd_path.startswith('\\') or lcd_path.startswith('/'):
                lcd_path = lcd_path.lstrip('\\/')
                MediLink_ConfigLoader.log("Removed leading slash from local storage path: {}".format(original_lcd_path), level="DEBUG")

            # Remove trailing slash if present
            if lcd_path.endswith('\\') or lcd_path.endswith('/'):
                lcd_path = lcd_path.rstrip('\\/')
                MediLink_ConfigLoader.log("Removed trailing slash from local storage path: {}".format(original_lcd_path), level="DEBUG")

            # Determine if quotes are necessary (e.g., if path contains spaces)
            if ' ' in lcd_path:
                lcd_command = 'lcd "{}"'.format(lcd_path)
                MediLink_ConfigLoader.log("Constructed lcd command with quotes: {}".format(lcd_command), level="DEBUG")
            else:
                lcd_command = 'lcd {}'.format(lcd_path)
                MediLink_ConfigLoader.log("Constructed lcd command without quotes: {}".format(lcd_command), level="DEBUG")

            command.append(lcd_command)

            # Handle filemask input
            if filemask:
                # TODO: Implement logic to translate filemask into WinSCP syntax
                # This should handle cases where filemask is a list, JSON, dictionary, or None.
                # Example: Convert to a string like "*.{ext1}|*.{ext2}|*.{ext3}".
                if isinstance(filemask, list):
                    filemask_str = '|'.join(['*.' + ext for ext in filemask])
                elif isinstance(filemask, dict):
                    filemask_str = '|'.join(['*.' + ext for ext in filemask.keys()])
                elif isinstance(filemask, str):
                    filemask_str = filemask  # Assume it's already in the correct format
                else:
                    filemask_str = '*'  # Default to all files if filemask is None or unsupported type
            else:
                filemask_str = '*'  # Default to all files if filemask is None

            # Use synchronize command for efficient downloading
            if newer_than:
                command.append('synchronize local -filemask="{}" -newerthan={}'.format(filemask_str, newer_than))
            else:
                command.append('synchronize local -filemask="{}"'.format(filemask_str))

        # Close and exit commands
        command += ['close', 'exit']
        MediLink_ConfigLoader.log("[Build Command] WinSCP command: {}".format(command))
        return command

    except Exception as e:
        MediLink_ConfigLoader.log("Error in build_command: {}. Reverting to original implementation.".format(e), level="ERROR")

        # Fallback to original implementation
        # Handle upload operation
        if operation_type == "upload":
            if not files:
                MediLink_ConfigLoader.log("Error: No files provided for upload operation.", level="ERROR")
                raise ValueError("No files provided for upload operation.")
            command.extend(["put {}".format(os.path.normpath(file_path)) for file_path in files])

        # Handle download operation
        else:
            command.append('get *')

        # Close and exit commands
        command.extend(['close', 'exit'])
        MediLink_ConfigLoader.log("[Build Command] Original WinSCP command: {}".format(command))
        return command

def simulate_operation(operation_type, files, config):
    MediLink_ConfigLoader.log("Test Mode is enabled! Simulating WinSCP {} operation.".format(operation_type))
    
    if operation_type == 'upload' and files:
        MediLink_ConfigLoader.log("Simulating 3 second delay for upload operation for files: {}".format(files))
        time.sleep(3)
        return [os.path.normpath(file) for file in files if os.path.exists(file)]
    elif operation_type == 'download':
        MediLink_ConfigLoader.log("Simulating 3 second delay for download operation. No files to download in test mode.")
        time.sleep(3)
        return []
    else:
        MediLink_ConfigLoader.log("Invalid operation type during simulation: {}".format(operation_type))
        return []

def execute_winscp_command(command, operation_type, files, local_storage_path):
    """
    Execute the WinSCP command for the specified operation type.
    """
    MediLink_ConfigLoader.log("Executing WinSCP command for operation type: {}".format(operation_type))
    
    try:
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
        stdout, stderr = process.communicate()
    except Exception as e:
        MediLink_ConfigLoader.log("Error occurred while executing WinSCP command: {}".format(e), level="ERROR")
        return []  # Return an empty list instead of None

    if process.returncode == 0:
        MediLink_ConfigLoader.log("WinSCP {} operation completed successfully.".format(operation_type))

        if operation_type == 'download':
            downloaded_files = list_downloaded_files(local_storage_path) # BUG This isn't behaving correctly because the local_storage_path isn't where winscp is dumping the files
            MediLink_ConfigLoader.log("Files currently located in local_storage_path: {}".format(downloaded_files), level="DEBUG")

            if not downloaded_files:
                MediLink_ConfigLoader.log("No files were downloaded or an error occurred during the listing process.", level="WARNING")
            return downloaded_files

        elif operation_type == 'upload':
            uploaded_files = [os.path.normpath(file) for file in files if os.path.exists(file)]
            MediLink_ConfigLoader.log("Uploaded files: {}".format(uploaded_files), level="DEBUG")
            return uploaded_files
    else:
        error_message = stderr.decode('utf-8').strip()
        MediLink_ConfigLoader.log("Failed to {} files. Exit code: {}. Details: {}".format(
            operation_type, process.returncode, error_message), level="ERROR")
        return []  # Return an empty list instead of None

def list_downloaded_files(local_storage_path):

    MediLink_ConfigLoader.log("Listing downloaded files in local storage path: {}".format(local_storage_path))
    
    # Initialize an empty list to hold file paths
    downloaded_files = []

    try:
        # Walk through the directory and collect all file paths
        for root, _, files in os.walk(local_storage_path):
            for file in files:
                file_path = os.path.join(root, file)
                downloaded_files.append(file_path)
                MediLink_ConfigLoader.log("File found: {}".format(file_path), level="DEBUG")
        
        if not downloaded_files:
            MediLink_ConfigLoader.log("No files found in the directory: {}".format(local_storage_path), level="WARNING")

    except Exception as e:
        MediLink_ConfigLoader.log("Error occurred while listing files in {}: {}".format(local_storage_path, e), level="ERROR")

    # Ensure that the function always returns a list
    return downloaded_files

def detect_new_files(directory_path, file_extension='.DAT'):
    """
    Scans the specified directory for new files with a given extension and adds a timestamp if needed.
    
    :param directory_path: Path to the directory containing files to be detected.
    :param file_extension: Extension of the files to detect.
    :return: A tuple containing a list of paths to new files detected in the directory and a flag indicating if a new file was just renamed.
    """
    MediLink_ConfigLoader.log("Scanning directory: {}".format(directory_path), level="INFO")
    detected_file_paths = []
    file_flagged = False
    
    try:
        filenames = os.listdir(directory_path)
        MediLink_ConfigLoader.log("Files in directory: {}".format(filenames), level="INFO")
        
        for filename in filenames:
            MediLink_ConfigLoader.log("Checking file: {}".format(filename), level="INFO")
            if filename.endswith(file_extension):
                MediLink_ConfigLoader.log("File matches extension: {}".format(file_extension), level="INFO")
                name, ext = os.path.splitext(filename)
                MediLink_ConfigLoader.log("File name: {}, File extension: {}".format(name, ext), level="INFO")
                
                if not is_timestamped(name):
                    MediLink_ConfigLoader.log("File is not timestamped: {}".format(filename), level="INFO")
                    new_name = "{}_{}{}".format(name, datetime.now().strftime('%Y%m%d_%H%M%S'), ext)
                    os.rename(os.path.join(directory_path, filename), os.path.join(directory_path, new_name))
                    MediLink_ConfigLoader.log("Renamed file from {} to {}".format(filename, new_name), level="INFO")
                    file_flagged = True
                    filename = new_name
                else:
                    MediLink_ConfigLoader.log("File is already timestamped: {}".format(filename), level="INFO")
                
                file_path = os.path.join(directory_path, filename)
                detected_file_paths.append(file_path)
                MediLink_ConfigLoader.log("Detected file path: {}".format(file_path), level="INFO")
    
    except Exception as e:
        MediLink_ConfigLoader.log("Error occurred: {}".format(str(e)), level="INFO")
    
    MediLink_ConfigLoader.log("Detected files: {}".format(detected_file_paths), level="INFO")
    MediLink_ConfigLoader.log("File flagged status: {}".format(file_flagged), level="INFO")
    
    return detected_file_paths, file_flagged

def is_timestamped(name):
    """
    Checks if the given filename has a timestamp in the expected format.
    
    :param name: The name of the file without extension.
    :return: True if the filename includes a timestamp, False otherwise.
    """
    # Regular expression to match timestamps in the format YYYYMMDD_HHMMSS
    timestamp_pattern = re.compile(r'.*_\d{8}_\d{6}$')
    return bool(timestamp_pattern.match(name))

def organize_patient_data_by_endpoint(detailed_patient_data):
    """
    Organizes detailed patient data by their confirmed endpoints.
    This simplifies processing and conversion per endpoint basis, ensuring that claims are generated and submitted
    according to the endpoint-specific requirements.

    :param detailed_patient_data: A list of dictionaries, each containing detailed patient data including confirmed endpoint.
    :return: A dictionary with endpoints as keys and lists of detailed patient data as values for processing.
    """
    organized = {}
    for data in detailed_patient_data:
        # Retrieve confirmed endpoint from each patient's data
        endpoint = data['confirmed_endpoint'] if 'confirmed_endpoint' in data else data['suggested_endpoint']
        # Initialize a list for the endpoint if it doesn't exist
        if endpoint not in organized:
            organized[endpoint] = []
        organized[endpoint].append(data)
    return organized

def confirm_all_suggested_endpoints(detailed_patient_data):
    """
    Confirms all suggested endpoints for each patient's detailed data.
    """
    for data in detailed_patient_data:
        if 'confirmed_endpoint' not in data:
            data['confirmed_endpoint'] = data['suggested_endpoint']
    return detailed_patient_data

def bulk_edit_insurance_types(detailed_patient_data, insurance_options):
    # Allow user to edit insurance types in a table-like format with validation
    print("Edit Insurance Type (Enter the 2-character code). Enter 'LIST' to display available insurance types.")

    for data in detailed_patient_data:
        current_insurance_type = data['insurance_type']
        current_insurance_description = insurance_options.get(current_insurance_type, "Unknown")
        print("({}) {:<25} | Current Ins. Type: {} - {}".format(
            data['patient_id'], data['patient_name'], current_insurance_type, current_insurance_description))

        while True:
            new_insurance_type = input("Enter new insurance type (or press Enter to keep current): ").upper()
            if new_insurance_type == 'LIST':
                MediLink_UI.display_insurance_options(insurance_options)
            elif not new_insurance_type or new_insurance_type in insurance_options:
                if new_insurance_type:
                    data['insurance_type'] = new_insurance_type
                break
            else:
                print("Invalid insurance type. Please enter a valid 2-character code or type 'LIST' to see options.")

def review_and_confirm_changes(detailed_patient_data, insurance_options):
    # Review and confirm changes
    print("\nReview changes:")
    print("{:<20} {:<10} {:<30}".format("Patient Name", "Ins. Type", "Description"))
    print("="*65)
    for data in detailed_patient_data:
        insurance_type = data['insurance_type']
        insurance_description = insurance_options.get(insurance_type, "Unknown")
        print("{:<20} {:<10} {:<30}".format(data['patient_name'], insurance_type, insurance_description))
    confirm = input("\nConfirm changes? (y/n): ").strip().lower()
    return confirm in ['y', 'yes', '']


# MediLink_Decoder.py
import os, sys, csv
from MediLink_ConfigLoader import load_configuration, log
from MediLink_Parser import parse_era_content, parse_277_content, parse_277IBR_content, parse_277EBR_content, parse_dpt_content, parse_ebt_content, parse_ibt_content

# Define new_fieldnames globally
new_fieldnames = ['Claim #', 'Payer', 'Status', 'Patient', 'Proc.', 'Serv.', 'Allowed', 'Paid', 'Pt Resp', 'Charged']

class UnifiedRecord:
    def __init__(self, claim_number='', status='', patient='', payer='', proc_date='', serv_date='', allowed='', paid='', pt_resp='', charged=''):
        self.claim_number = claim_number
        self.payer = payer  # Added payer to the constructor
        self.status = status
        self.patient = patient
        self.proc_date = proc_date
        self.serv_date = serv_date
        self.allowed = allowed
        self.paid = paid
        self.pt_resp = pt_resp
        self.charged = charged

    def to_dict(self):
        return {
            'Claim #': self.claim_number,
            'Payer': self.payer,  # Added payer to the dictionary representation
            'Status': self.status,
            'Patient': self.patient,
            'Proc.': self.proc_date,
            'Serv.': self.serv_date,
            'Allowed': self.allowed,
            'Paid': self.paid,
            'Pt Resp': self.pt_resp,
            'Charged': self.charged
        }

    def __repr__(self):
        return ("UnifiedRecord(claim_number='{0}', status='{1}', patient='{2}', payer='{3}', proc_date='{4}', serv_date='{5}', "
                "allowed='{6}', paid='{7}', pt_resp='{8}', charged='{9}')").format(
            self.claim_number, self.status, self.patient, self.payer, self.proc_date,
            self.serv_date, self.allowed, self.paid, self.pt_resp, self.charged)  # Added payer to the repr

def process_decoded_file(file_path, output_directory, return_records=False, debug=False): # Renamed from process_file
    os.makedirs(output_directory, exist_ok=True)

    file_type = determine_file_type(file_path)
    content = read_file(file_path)

    parse_functions = {
        'ERA': parse_era_content,
        '277': parse_277_content,
        '277IBR': parse_277IBR_content,
        '277EBR': parse_277EBR_content,
        'DPT': parse_dpt_content,
        'EBT': parse_ebt_content,
        'IBT': parse_ibt_content
    }

    parse_function = parse_functions.get(file_type)
    if parse_function is None:
        log("Unsupported file type: {}".format(file_type))
        return []

    records = parse_function(content, debug=debug)
    formatted_records = format_records(records, file_type)

    if not return_records:
        display_table([record.to_dict() for record in formatted_records])
        output_file_path = os.path.join(output_directory, "{}_decoded.csv".format(os.path.basename(file_path)))
        write_records_to_csv(formatted_records, output_file_path)
        log("Decoded data written to {}".format(output_file_path))

    return formatted_records  # Returns list of UnifiedRecord instances

def determine_file_type(file_path):
    file_extensions = {
        '.era': 'ERA',
        '.277': '277',
        '.277ibr': '277IBR',
        '.277ebr': '277EBR',
        '.dpt': 'DPT',
        '.ebt': 'EBT',
        '.ibt': 'IBT'
    }
    
    for ext, file_type in file_extensions.items():
        if file_path.endswith(ext):
            return file_type
            
    log("Unsupported file type for file: {}".format(file_path))
    return None

def read_file(file_path):
    with open(file_path, 'r') as file:
        content = file.read()
    return content

def write_records_to_csv(records, output_file_path):
    if not records:
        log("No records to write.", 'error')
        return

    # Use the global variable for fieldnames
    fieldnames = new_fieldnames
    
    with open(output_file_path, 'w', newline='') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()
        for record in records:
            writer.writerow(record.to_dict())

def format_records(records, file_type):
    formatted_records = []
    seen_claim_numbers = set()  # Set to track unique claim numbers

    for i, record in enumerate(records):
        # Determine the claim number based on the file type
        if file_type == 'IBT':
            claim_number = record.get('Patient Control Number', '')
        elif file_type == 'ERA':
            claim_number = record.get('Chart Number', '')
        elif file_type == 'EBT':
            claim_number = record.get('Patient Control Number', '')
        else:
            claim_number = ''  # Default to empty if file type is not recognized

        # Skip records without a claim number
        if not claim_number:
            log("Record {} missing claim_number. Skipping.".format(i + 1), level="WARNING")
            continue

        # Check for duplicates
        if claim_number in seen_claim_numbers:
            log("Duplicate claim_number {} found at record {}. Skipping.".format(claim_number, i + 1), level="DEBUG")
            continue

        seen_claim_numbers.add(claim_number)  # Add claim number to the set. BUG This isn't successfully removing duplicate records. -- it might be across files and not within them.

        unified_record = UnifiedRecord()

        # Populate the unified_record based on the file type
        if file_type == 'IBT':
            unified_record.claim_number = claim_number
            unified_record.status = record.get('Status', '')
            unified_record.patient = record.get('Patient Name', '')
            unified_record.proc_date = format_date(record.get('To Date', ''))
            unified_record.serv_date = format_date(record.get('From Date', ''))
            unified_record.charged = record.get('Charge', '')

        elif file_type == 'ERA':
            unified_record.claim_number = claim_number
            unified_record.status = record.get('claimStatus', '')
            unified_record.patient = record.get('Patient', '')
            unified_record.proc_date = format_date(record.get('processed_date', ''))
            unified_record.serv_date = format_date(record.get('Date of Service', ''))
            unified_record.allowed = record.get('Allowed Amount', '')
            unified_record.paid = record.get('Amount Paid', '')
            unified_record.pt_resp = record.get('Patient Responsibility', '')
            unified_record.charged = record.get('Charge', '')

        elif file_type == 'EBT':
            if 'Patient Control Number' in record:
                unified_record.claim_number = claim_number
                message_type = record.get('Message Type', '').upper()
                status_mapping = {
                    'A': 'Accepted',
                    'R': 'Rejected',
                }
                # unified_record.status = status_mapping.get(message_type, message_type)
                unified_record.status = record.get('Message', '')
                unified_record.payer = record.get('Message Initiator', '')
                unified_record.patient = record.get('Patient Name', '')
                unified_record.proc_date = format_date(record.get('To Date', ''))
                unified_record.serv_date = format_date(record.get('From Date', ''))
                unified_record.allowed = ''
                unified_record.paid = ''
                unified_record.pt_resp = ''
                unified_record.charged = record.get('Charge', '')
                log("Formatted EBT Record {}: {}".format(i + 1, unified_record), level="DEBUG")
            else:
                log("Skipped non-claim EBT Record {}: {}".format(i + 1, record), level="DEBUG")
                continue

        # Append the unified record to the list
        formatted_records.append(unified_record)

    return formatted_records

def format_date(date_str):
    if date_str and len(date_str) >= 8:
        return date_str[4:6] + '-' + date_str[6:8]  # Adjusted to match sample date format 'YYYYMMDD'
    return ''

def display_table(records):
    """
    Display records in a formatted table after deduplication.
    
    Args:
        records (list): List of UnifiedRecord instances.
    """
    # Deduplicate records before displaying
    records = deduplicate_records(records)

    if not records:
        print("No records to display.")
        return

    # Determine which fields have at least one non-empty value
    used_fields = [field for field in new_fieldnames if any(str(record.get(field, '')).strip() for record in records)]

    if not used_fields:
        print("No data to display.")
        return

    # Calculate column widths based on the longest item in each used column
    col_widths = {field: len(field) for field in used_fields}

    for record in records:
        for field in used_fields:
            col_widths[field] = max(col_widths[field], len(str(record.get(field, ''))))

    # Create table header
    header = " | ".join("{:<{}}".format(field, col_widths[field]) for field in used_fields)
    print(header)
    print("-" * len(header))

    # Create table rows
    for record in records:
        row = " | ".join("{:<{}}".format(str(record.get(field, '')), col_widths[field]) for field in used_fields)
        print(row)

def display_consolidated_records(records):
    """
    Display the consolidated records in a formatted table.
    Removes any records that are completely empty or only contain whitespace.
    If no valid records are found, displays a message to that effect.
    """
    # Deduplicate records before displaying
    records = deduplicate_records(records)

    # If records are UnifiedRecord instances, convert them to dictionaries
    if records and isinstance(records[0], UnifiedRecord):
        dict_records = [record.to_dict() for record in records]
    elif records and isinstance(records[0], dict):
        dict_records = records
    else:
        log("Invalid record format for display.", level="ERROR")
        return
    
    # Filter out records that are completely empty or only contain whitespace
    filtered_records = [
        record for record in dict_records 
        if any(str(record.get(field, '')).strip() for field in new_fieldnames)
    ]
    
    if not filtered_records:
        print("No valid records to display after filtering empty rows.")
        return
    
    # Determine which fields have at least one non-empty value
    used_fields = [field for field in new_fieldnames if any(str(record.get(field, '')).strip() for record in filtered_records)]

    if not used_fields:
        print("No data to display.")
        return
    
    # Calculate column widths based on the longest item in each used column
    col_widths = {field: len(field) for field in used_fields}
    
    for record in filtered_records:
        for field in used_fields:
            col_widths[field] = max(col_widths[field], len(str(record.get(field, ''))))
    
    # Print header
    header = " | ".join("{:<{}}".format(field, col_widths[field]) for field in used_fields)
    print(header)
    print("-" * len(header))
    
    # Print each row
    for record in filtered_records:
        row = " | ".join("{:<{}}".format(str(record.get(field, '')), col_widths[field]) for field in used_fields)
        print(row)

def deduplicate_records(records):
    """
    Remove duplicate records based on claim_number.
    
    Args:
        records (list): List of UnifiedRecord instances.
    
    Returns:
        list: List of unique UnifiedRecord instances.
    """
    unique_records_dict = {}
    for record in records:
        if record.claim_number not in unique_records_dict:
            unique_records_dict[record.claim_number] = record
        else:
            log("Duplicate record found for claim_number {}. Skipping.".format(record.claim_number), "DEBUG")
    
    return list(unique_records_dict.values())

if __name__ == "__main__":
    config, _ = load_configuration()
    
    files = sys.argv[1:]
    if not files:
        log("No files provided as arguments.", 'error')
        sys.exit(1)

    output_directory = config['MediLink_Config'].get('local_storage_path')
    all_records = []
    for file_path in files:
        try:
            records = process_decoded_file(file_path, output_directory, return_records=True)
            all_records.extend(records)
        except Exception as e:
            log("Failed to process {}: {}".format(file_path, e), 'error')
    
    # Call the deduplication function
    unique_records = deduplicate_records(all_records)

    display_consolidated_records([record.to_dict() for record in unique_records])

    if input("Do you want to export the consolidated records to a CSV file? (y/n): ").strip().lower() == 'y':
        consolidated_csv_path = os.path.join(output_directory, "Consolidated_Records.csv")
        write_records_to_csv(unique_records, consolidated_csv_path)
        log("Consolidated records written to {}".format(consolidated_csv_path))



# MediLink_Gmail.py
import sys, os, subprocess, time, webbrowser, requests, json, ssl, signal
from MediLink_ConfigLoader import log, load_configuration
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread, Event
import platform
import ctypes

config, _ = load_configuration()
local_storage_path = config['MediLink_Config']['local_storage_path']
downloaded_emails_file = os.path.join(local_storage_path, 'downloaded_emails.txt')

server_port = 8000
cert_file = 'server.cert'
key_file = 'server.key'
openssl_cnf = 'MediLink\\openssl.cnf' # This file needs to be located in the same place as where MediCafe is run from (so MediBot folder?)

httpd = None  # Global variable for the HTTP server
shutdown_event = Event()  # Event to signal shutdown

# Define the scopes for the Gmail API and other required APIs
SCOPES = ' '.join([
    'https://www.googleapis.com/auth/gmail.modify',
    'https://www.googleapis.com/auth/gmail.compose',
    'https://www.googleapis.com/auth/gmail.readonly',
    'https://www.googleapis.com/auth/script.external_request',
    'https://www.googleapis.com/auth/userinfo.email',
    'https://www.googleapis.com/auth/script.scriptapp',
    'https://www.googleapis.com/auth/drive'
])

# Path to token.json file
TOKEN_PATH = 'token.json'

# Determine the operating system and version
os_name = platform.system()
os_version = platform.release()

# Set the credentials path based on the OS and version
if os_name == 'Windows' and 'XP' in os_version:
    CREDENTIALS_PATH = 'F:\\Medibot\\json\\credentials.json'
else:
    CREDENTIALS_PATH = 'json\\credentials.json'

# Log the selected path for verification
log("Using CREDENTIALS_PATH: {}".format(CREDENTIALS_PATH), config, level="INFO")

REDIRECT_URI = 'https://127.0.0.1:8000'

def get_authorization_url():
    with open(CREDENTIALS_PATH, 'r') as credentials_file:
        credentials = json.load(credentials_file)
    client_id = credentials['web']['client_id']
    auth_url = (
        "https://accounts.google.com/o/oauth2/v2/auth?"
        "response_type=code&"
        "client_id={}&"
        "redirect_uri={}&"
        "scope={}&"
        "access_type=offline&"  # Requesting offline access allows the application to obtain a refresh token, enabling it to access resources even when the user is not actively using the app. This is useful for long-lived sessions.
        # To improve user experience, consider changing this to 'online' if you don't need offline access:
        # "access_type=online&"  # Use this if you only need access while the user is actively using the app and don't require a refresh token.
        
        "prompt=consent"  # This forces the user to re-consent to the requested scopes every time they authenticate. While this is useful for ensuring the user is aware of the permissions being granted, it can be modified to 'none' or omitted entirely if the application is functioning correctly and tokens are being refreshed properly.
        # To improve user experience, consider changing this to 'none' if you want to avoid showing the consent screen every time:
        # "prompt=none"  # Use this if you want to skip the consent screen for users who have already granted permissions.
        # Alternatively, you can omit the prompt parameter entirely to use the default behavior:
        # # "prompt="  # Omitting this will show the consent screen only when necessary.
    ).format(client_id, REDIRECT_URI, SCOPES)
    log("Generated authorization URL: {}".format(auth_url))
    return auth_url

def exchange_code_for_token(auth_code, retries=3):
    for attempt in range(retries):
        try:
            with open(CREDENTIALS_PATH, 'r') as credentials_file:
                credentials = json.load(credentials_file)
            token_url = "https://oauth2.googleapis.com/token"
            data = {
                'code': auth_code,
                'client_id': credentials['web']['client_id'],
                'client_secret': credentials['web']['client_secret'],
                'redirect_uri': REDIRECT_URI,
                'grant_type': 'authorization_code'
            }
            response = requests.post(token_url, data=data)
            log("Token exchange response: Status code {}, Body: {}".format(response.status_code, response.text))
            token_response = response.json()
            if response.status_code == 200:
                token_response['token_time'] = time.time()
                return token_response
            else:
                log("Token exchange failed: {}".format(token_response))
                if attempt < retries - 1:
                    log("Retrying token exchange... (Attempt {}/{})".format(attempt + 1, retries))
        except Exception as e:
            log("Error during token exchange: {}".format(e))
    return {}

def get_access_token():
    if os.path.exists(TOKEN_PATH):
        with open(TOKEN_PATH, 'r') as token_file:
            token_data = json.load(token_file)
            log("Loaded token data:\n {}".format(token_data))
            
            if 'access_token' in token_data and 'expires_in' in token_data:
                try:
                    # Use current time if 'token_time' is missing
                    token_time = token_data.get('token_time', time.time())
                    token_expiry_time = token_time + token_data['expires_in']
                
                except KeyError as e:
                    log("KeyError while accessing token data: {}".format(e))
                    return None
                
                if token_expiry_time > time.time():
                    log("Access token is still valid. Expires in {} seconds.".format(token_expiry_time - time.time()))
                    return token_data['access_token']
                else:
                    log("Access token has expired. Current time: {}, Expiry time: {}".format(time.time(), token_expiry_time))
                    new_token_data = refresh_access_token(token_data.get('refresh_token'))
                    if 'access_token' in new_token_data:
                        new_token_data['token_time'] = time.time()
                        with open(TOKEN_PATH, 'w') as token_file:
                            json.dump(new_token_data, token_file)
                        log("Access token refreshed successfully. New token data: {}".format(new_token_data))
                        return new_token_data['access_token']
                    else:
                        log("Failed to refresh access token. New token data: {}".format(new_token_data))
                        return None
    log("Access token not found. Please authenticate.")
    return None

def refresh_access_token(refresh_token):
    log("Refreshing access token.")
    with open(CREDENTIALS_PATH, 'r') as credentials_file:
        credentials = json.load(credentials_file)
    token_url = "https://oauth2.googleapis.com/token"
    data = {
        'client_id': credentials['web']['client_id'],
        'client_secret': credentials['web']['client_secret'],
        'refresh_token': refresh_token,
        'grant_type': 'refresh_token'
    }
    response = requests.post(token_url, data=data)
    log("Refresh token response: Status code {}, Body:\n {}".format(response.status_code, response.text))
    if response.status_code == 200:
        log("Access token refreshed successfully.")
        return response.json()
    else:
        log("Failed to refresh access token. Status code: {}".format(response.status_code))
        return {}

def bring_window_to_foreground():
    """Brings the current window to the foreground on Windows."""
    try:
        if platform.system() == 'Windows':
            # Get the current process ID
            pid = os.getpid()
            # Get the window handle for the current process
            hwnd = ctypes.windll.user32.GetForegroundWindow()
            # Get the process ID of the window
            current_pid = ctypes.c_ulong()
            ctypes.windll.user32.GetWindowThreadProcessId(hwnd, ctypes.byref(current_pid))
            
            # If the window is not ours, try to bring it to front
            if current_pid.value != pid:
                # Try to set the window to foreground
                ctypes.windll.user32.SetForegroundWindow(hwnd)
                # If that fails, try the alternative method
                if ctypes.windll.user32.GetForegroundWindow() != hwnd:
                    ctypes.windll.user32.ShowWindow(hwnd, 9)  # SW_RESTORE = 9
                    ctypes.windll.user32.SetForegroundWindow(hwnd)
    except Exception as e:
        log("Error bringing window to foreground: {}".format(e))

class RequestHandler(BaseHTTPRequestHandler):
    def _set_headers(self):
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'POST, OPTIONS')
        self.send_header('Access-Control-Allow-Headers', 'Content-Type')
        self.send_header('Content-type', 'application/json')

    def do_OPTIONS(self):
        self.send_response(200)
        self._set_headers()
        self.end_headers()

    def do_POST(self):
        if self.path == '/download':
            content_length = int(self.headers['Content-Length'])
            post_data = self.rfile.read(content_length)
            data = json.loads(post_data.decode('utf-8'))
            links = data.get('links', [])

            # Log the content of links
            log("Received links: {}".format(links))

            file_ids = [link.get('fileId', None) for link in links if link.get('fileId')]
            log("File IDs received from client: {}".format(file_ids))
            
            # Proceed with downloading files
            download_docx_files(links)
            self.send_response(200)
            self._set_headers()  # Include CORS headers
            self.end_headers()
            response = json.dumps({"status": "success", "message": "All files downloaded", "fileIds": file_ids})
            self.wfile.write(response.encode('utf-8'))
            shutdown_event.set()
            bring_window_to_foreground()  # Bring window to foreground after download
        elif self.path == '/shutdown':
            log("Shutdown request received.")
            self.send_response(200)
            self._set_headers()
            self.end_headers()
            response = json.dumps({"status": "success", "message": "Server is shutting down."})
            self.wfile.write(response.encode('utf-8'))
            shutdown_event.set()  # Signal shutdown event instead of calling stop_server directly
        elif self.path == '/delete-files':
            content_length = int(self.headers['Content-Length'])
            post_data = self.rfile.read(content_length)
            data = json.loads(post_data.decode('utf-8'))
            file_ids = data.get('fileIds', [])
            log("File IDs to delete received from client: {}".format(file_ids))
            
            if not isinstance(file_ids, list):
                self.send_response(400)
                self._set_headers()
                self.end_headers()
                response = json.dumps({"status": "error", "message": "Invalid fileIds parameter."})
                self.wfile.write(response.encode('utf-8'))
                return
            
            self.send_response(200)
            self._set_headers()  # Include CORS headers
            self.end_headers()
            response = json.dumps({"status": "success", "message": "Files deleted successfully."})
            self.wfile.write(response.encode('utf-8'))
        else:
            self.send_response(404)
            self.end_headers()

    def do_GET(self):
        log("Full request path: {}".format(self.path))  # Log the full path for debugging
        if self.path.startswith("/?code="):
            auth_code = self.path.split('=')[1].split('&')[0]
            auth_code = requests.utils.unquote(auth_code)  # Decode if URL-encoded
            log("Received authorization code: {}".format(auth_code))
            if is_valid_authorization_code(auth_code):
                try:
                    token_response = exchange_code_for_token(auth_code)
                    if 'access_token' not in token_response:
                        # Check for specific error message
                        if token_response.get("status") == "error":
                            self.send_response(400)
                            self.send_header('Content-type', 'text/html')
                            self.end_headers()
                            self.wfile.write(token_response["message"].encode())
                            return
                        # Handle other cases
                        raise ValueError("Access token not found in response.")
                except Exception as e:
                    log("Error during token exchange: {}".format(e))
                    self.send_response(500)
                    self.send_header('Content-type', 'text/html')
                    self.end_headers()
                    self.wfile.write("An error occurred during authentication. Please try again.".encode())
                else:
                    log("Token response: {}".format(token_response))  # Add this line
                    if 'access_token' in token_response:
                        with open(TOKEN_PATH, 'w') as token_file:
                            json.dump(token_response, token_file)
                        self.send_response(200)
                        self.send_header('Content-type', 'text/html')
                        self.end_headers()
                        self.wfile.write("Authentication successful. You can close this window now.".encode())
                        initiate_link_retrieval(config)  # Pass config here
                    else:
                        log("Authentication failed with response: {}".format(token_response))  # Log the full response
                        if 'error' in token_response:
                            error_description = token_response.get('error_description', 'No description provided.')
                            log("Error details: {}".format(error_description))  # Log specific error details

                        # Provide user feedback based on the error
                        if token_response.get('error') == 'invalid_grant':
                            log("Invalid grant error encountered. Authorization code: {}, Response: {}".format(auth_code, token_response))
                            check_invalid_grant_causes(auth_code)
                            clear_token_cache()  # Clear the cache on invalid grant
                            user_message = "Authentication failed: Invalid or expired authorization code. Please try again."
                        else:
                            user_message = "Authentication failed. Please check the logs for more details."

                        self.send_response(400)
                        self.send_header('Content-type', 'text/html')
                        self.end_headers()
                        self.wfile.write(user_message.encode())
                        shutdown_event.set()  # Signal shutdown event after failed authentication
            else:
                log("Invalid authorization code format: {}".format(auth_code))
                self.send_response(400)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.wfile.write("Invalid authorization code format. Please try again.".encode())
                shutdown_event.set()  # Signal shutdown event after failed authentication
        elif self.path == '/downloaded-emails':
            self.send_response(200)
            self._set_headers()
            self.end_headers()
            downloaded_emails = load_downloaded_emails()
            response = json.dumps({"downloadedEmails": list(downloaded_emails)})
            self.wfile.write(response.encode('utf-8'))
        else:
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
            self.send_header('Access-Control-Allow-Headers', 'Content-Type')
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write(b'HTTPS server is running.')

def generate_self_signed_cert(cert_file, key_file):
    log("Checking if certificate file exists: " + cert_file)
    log("Checking if key file exists: " + key_file)
    if not os.path.exists(cert_file) or not os.path.exists(key_file):
        log("Generating self-signed SSL certificate...")
        cmd = [
            'openssl', 'req', '-config', openssl_cnf, '-nodes', '-new', '-x509',
            '-keyout', key_file,
            '-out', cert_file,
            '-days', '365'
            #'-subj', '/C=US/ST=...' The openssl.cnf file contains default values for these fields, but they can be overridden by the -subj option.
        ]
        try:
            log("Running command: " + ' '.join(cmd))
            result = subprocess.call(cmd)
            log("Command finished with result: " + str(result))
            if result != 0:
                raise RuntimeError("Failed to generate self-signed certificate")
            log("Self-signed SSL certificate generated.")
        except Exception as e:
            log("Error generating self-signed certificate: {}".format(e))
            raise

def run_server():
    global httpd
    try:
        log("Attempting to start server on port " + str(server_port))
        server_address = ('0.0.0.0', server_port)  # Bind to all interfaces
        httpd = HTTPServer(server_address, RequestHandler)
        log("Attempting to wrap socket with SSL. cert_file=" + cert_file + ", key_file=" + key_file)
        
        if not os.path.exists(cert_file):
            log("Error: Certificate file not found: " + cert_file)
        if not os.path.exists(key_file):
            log("Error: Key file not found: " + key_file)
        
        httpd.socket = ssl.wrap_socket(httpd.socket, certfile=cert_file, keyfile=key_file, server_side=True)
        log("Starting HTTPS server on port {}".format(server_port))
        httpd.serve_forever()
    except Exception as e:
        log("Error in serving: {}".format(e))
        stop_server()

def stop_server():
    global httpd
    if httpd:
        log("Stopping HTTPS server.")
        httpd.shutdown()
        httpd.server_close()
        log("HTTPS server stopped.")
    shutdown_event.set()  # Signal shutdown event
    bring_window_to_foreground()  # Bring window to foreground after shutdown

def load_downloaded_emails():
    downloaded_emails = set()
    if os.path.exists(downloaded_emails_file):
        with open(downloaded_emails_file, 'r') as file:
            downloaded_emails = set(line.strip() for line in file)
    log("Loaded downloaded emails: {}".format(downloaded_emails))
    return downloaded_emails

def download_docx_files(links):
    # Load the set of downloaded emails
    # TODO Test if any of these have a .csv extension and then move those to the right location locally.
    downloaded_emails = load_downloaded_emails()

    for link in links:
        try:
            url = link.get('url', '')
            filename = link.get('filename', '')
            
            # Log the variables to debug
            log("Processing link: url='{}', filename='{}'".format(url, filename))
            
            # Skip if email already downloaded
            if filename in downloaded_emails:
                log("Skipping already downloaded email: {}".format(filename))
                continue

            log("Downloading .docx file from URL: {}".format(url))
            response = requests.get(url, verify=False)  # Set verify to False for self-signed certs
            if response.status_code == 200:
                file_path = os.path.join(local_storage_path, filename)
                with open(file_path, 'wb') as file:
                    file.write(response.content)
                log("Downloaded .docx file: {}".format(filename))
                # Add to the set and save the updated list
                downloaded_emails.add(filename)
                with open(downloaded_emails_file, 'a') as file:
                    file.write(filename + '\n')
            else:
                log("Failed to download .docx file from URL: {}. Status code: {}".format(url, response.status_code))
        except Exception as e:
            log("Error downloading .docx file from URL: {}. Error: {}".format(url, e))

def open_browser_with_executable(url, browser_path=None):
    try:
        if browser_path:
            log("Attempting to open URL with provided executable: {} {}".format(browser_path, url))
            process = subprocess.Popen([browser_path, url], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()
            if process.returncode == 0:
                log("Browser opened with provided executable path using subprocess.Popen.")
            else:
                log("Browser failed to open using subprocess.Popen. Return code: {}. Stderr: {}".format(process.returncode, stderr))
        else:
            log("No browser path provided. Attempting to open URL with default browser: {}".format(url))
            webbrowser.open(url)
            log("Default browser opened.")
    except Exception as e:
        log("Failed to open browser: {}".format(e))

def initiate_link_retrieval(config):
    log("Initiating browser via implicit GET.")
    url_get = "https://script.google.com/macros/s/{}/exec?action=get_link".format(config['MediLink_Config']['webapp_deployment_id'])  # Use config here
    open_browser_with_executable(url_get)
    
    log("Preparing POST call.")
    url = "https://script.google.com/macros/s/{}/exec".format(config['MediLink_Config']['webapp_deployment_id'])  # Use config here
    downloaded_emails = list(load_downloaded_emails())
    payload = {
        "downloadedEmails": downloaded_emails
    }

    access_token = get_access_token()
    if not access_token:
        log("Access token not found. Please authenticate first.")
        shutdown_event.set()  # Signal shutdown event if token is not found
        return

    # Inspect the token to check its validity and permissions
    token_info = inspect_token(access_token)
    if token_info is None:
        log("Access token is invalid. Please re-authenticate.")
        shutdown_event.set()  # Signal shutdown event if token is invalid
        return

    # Proceed with the rest of the function if the token is valid
    headers = {
        'Authorization': 'Bearer {}'.format(access_token),
        'Content-Type': 'application/json'
    }
    
    log("Request headers: {}".format(headers))
    log("Request payload: {}".format(payload))

    handle_post_response(url, payload, headers)
        
def handle_post_response(url, payload, headers):
    try:
        response = requests.post(url, json=payload, headers=headers)
        log("Response status code: {}".format(response.status_code))
        log("Response body: {}".format(response.text))

        if response.status_code == 200:
            response_data = response.json()
            log("Parsed response data: {}".format(response_data))  # Log the parsed response data
            if response_data.get("status") == "error":
                log("Error message from server: {}".format(response_data.get("message")))
                print("Error: {}".format(response_data.get("message")))
                shutdown_event.set()  # Signal shutdown event after error
            else:
                log("Link retrieval initiated successfully.")
        elif response.status_code == 401:
            log("Unauthorized. Check if the token has the necessary scopes.Response body: {}".format(response.text))
            # Inspect the token to log its details
            token_info = inspect_token(headers['Authorization'].split(' ')[1])
            log("Token details: {}".format(token_info))
            shutdown_event.set()
        elif response.status_code == 403:
            log("Forbidden access. Ensure that the OAuth client has the correct permissions. Response body: {}".format(response.text))
            shutdown_event.set()
        elif response.status_code == 404:
            log("Not Found. Verify the URL and ensure the Apps Script is deployed correctly. Response body: {}".format(response.text))
            shutdown_event.set()
        else:
            log("Failed to initiate link retrieval. Unexpected status code: {}. Response body: {}".format(response.status_code, response.text))
            shutdown_event.set()
    except requests.exceptions.RequestException as e:
        log("RequestException during link retrieval initiation: {}".format(e))
        shutdown_event.set()
    except Exception as e:
        log("Unexpected error during link retrieval initiation: {}".format(e))
        shutdown_event.set()

def inspect_token(access_token):
    info_url = "https://www.googleapis.com/oauth2/v1/tokeninfo?access_token={}".format(access_token)
    try:
        response = requests.get(info_url)
        log("Token info: Status code {}, Body: {}".format(response.status_code, response.text))
        
        if response.status_code == 200:
            return response.json()
        else:
            log("Failed to inspect token. Status code: {}, Body: {}".format(response.status_code, response.text))
            # Check for invalid token
            if response.status_code == 400 and "invalid_token" in response.text:
                log("Access token is invalid. Deleting token.json and stopping the server.")
                delete_token_file()  # Delete the token.json file
                print("Access token is invalid. Please re-authenticate and restart the server.")
                stop_server()  # Stop the server
                return None  # Return None for invalid tokens
            return None  # Return None for other invalid tokens
    except Exception as e:
        log("Exception during token inspection: {}".format(e))
        return None

def delete_token_file():
    try:
        if os.path.exists(TOKEN_PATH):
            os.remove(TOKEN_PATH)
            log("Deleted token.json successfully.")
        else:
            log("token.json does not exist.")
    except Exception as e:
        log("Error deleting token.json: {}".format(e))

def signal_handler(sig, frame):
    log("Signal received: {}. Initiating shutdown.".format(sig))
    stop_server()
    sys.exit(0)

def auth_and_retrieval():
    access_token = get_access_token()
    if not access_token:
        log("Access token not found or expired. Please authenticate first.")
        #print("If the browser does not open automatically, please open the following URL in your browser to authorize the application:")
        auth_url = get_authorization_url()
        #print(auth_url)
        open_browser_with_executable(auth_url)
        shutdown_event.wait()  # Wait for the shutdown event to be set after authentication
    else:
        log("Access token found. Proceeding.")
        initiate_link_retrieval(config)  # Pass config here
        shutdown_event.wait()  # Wait for the shutdown event to be set

def is_valid_authorization_code(auth_code):
    # Check if the authorization code is not None and is a non-empty string
    if auth_code and isinstance(auth_code, str) and len(auth_code) > 0:  # Check for non-empty string
        return True
    log("Invalid authorization code format: {}".format(auth_code))
    return False

def clear_token_cache():
    if os.path.exists(TOKEN_PATH):
        os.remove(TOKEN_PATH)
        log("Cleared token cache.")

def check_invalid_grant_causes(auth_code):
    # TODO Implement this function in the future to check for common causes of invalid_grant error
    # Log potential causes for invalid_grant
    log("FUTURE IMPLEMENTATION: Checking common causes for invalid_grant error with auth code: {}".format(auth_code))
    # Example checks (you can expand this based on your needs)
"""
if is_code_used(auth_code):
        log("Authorization code has already been used.")
    if not is_redirect_uri_correct():
        log("Redirect URI does not match the registered URI.")
"""

if __name__ == "__main__":
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    try:
        # Generate SSL certificate if it doesn't exist
        generate_self_signed_cert(cert_file, key_file)

        from threading import Thread
        log("Starting server thread.")
        server_thread = Thread(target=run_server)
        server_thread.daemon = True
        server_thread.start()

        auth_and_retrieval()

        log("Stopping HTTPS server.")
        stop_server()  # Ensure the server is stopped
        log("Waiting for server thread to finish.")
        server_thread.join()  # Wait for the server thread to finish
    except KeyboardInterrupt:
        log("KeyboardInterrupt received, stopping server.")
        stop_server()
        sys.exit(0)
    except Exception as e:
        log("An error occurred: {}".format(e))
        stop_server()
        sys.exit(1)


# MediLink_ConfigLoader.py
import os, json, logging, sys, platform, yaml
from datetime import datetime
from collections import OrderedDict

"""
This function should be generalizable to have a initialization script over all the Medi* functions
"""
def load_configuration(config_path=os.path.join(os.path.dirname(__file__), '..', 'json', 'config.json'), crosswalk_path=os.path.join(os.path.dirname(__file__), '..', 'json', 'crosswalk.json')):
    """
    Loads endpoint configuration, credentials, and other settings from JSON or YAML files.
        
    Returns: A tuple containing dictionaries with configuration settings for the main config and crosswalk.
    """
    # TODO (Low Config Upgrade) The Medicare / Private differentiator flag probably needs to be pulled or passed to this.
    # BUG Hardcode sucks. This should probably be some local env variable. 
    # Detect the operating system
    if platform.system() == 'Windows' and platform.release() == 'XP':
        # Use F: paths for Windows XP
        config_path = "F:\\Medibot\\json\\config.json"
        crosswalk_path = "F:\\Medibot\\json\\crosswalk.json"
    else:
        # Use G: paths for other versions of Windows
        config_path = "G:\\My Drive\\Codes\\MediCafe\\json\\config.json"
        crosswalk_path = "G:\\My Drive\\Codes\\MediCafe\\json\\crosswalk.json"
    
    try:
        with open(config_path, 'r') as config_file:
            if config_path.endswith('.yaml') or config_path.endswith('.yml'):
                config = yaml.safe_load(config_file)
            elif config_path.endswith('.json'):
                config = json.load(config_file, object_pairs_hook=OrderedDict)
            else:
                raise ValueError("Unsupported configuration format.")
            
            if 'MediLink_Config' not in config:
                raise KeyError("MediLink_Config key is missing from the loaded configuration.")

        with open(crosswalk_path, 'r') as crosswalk_file:
            crosswalk = json.load(crosswalk_file)

        return config, crosswalk
    except ValueError as e:
        if isinstance(e, UnicodeDecodeError):
            print("Error decoding file: {}".format(e))
        else:
            print("Error parsing file: {}".format(e))
        sys.exit(1)  # Exit the script due to a critical error in configuration loading
    except FileNotFoundError:
        print("One or both configuration files not found. Config: {}, Crosswalk: {}".format(config_path, crosswalk_path))
        sys.exit(1)  # Exit the script due to a critical error in configuration loading
    except KeyError as e:
        print("Critical configuration is missing: {}".format(e))
        sys.exit(1)  # Exit the script due to a critical error in configuration loading
    except Exception as e:
        print("An unexpected error occurred while loading the configuration: {}".format(e))
        sys.exit(1)  # Exit the script due to a critical error in configuration loading
        
# Logs messages with optional error type and claim data.
def log(message, config=None, level="INFO", error_type=None, claim=None, verbose=False):
    
    # If config is not provided, load it
    if config is None:
        config, _ = load_configuration()
    
    # Setup logger if not already configured
    if not logging.root.handlers:
        local_storage_path = config['MediLink_Config'].get('local_storage_path', '.') if isinstance(config, dict) else '.'
        log_filename = datetime.now().strftime("Log_%m%d%Y.log")
        log_filepath = os.path.join(local_storage_path, log_filename)
        
        # Set logging level based on verbosity
        logging_level = logging.DEBUG if verbose else logging.INFO
        
        logging.basicConfig(level=logging_level,
                            format='%(asctime)s - %(levelname)s - %(message)s',
                            filename=log_filepath,
                            filemode='a')
    
    # Prepare log message
    claim_data = " - Claim Data: {}".format(claim) if claim else ""
    error_info = " - Error Type: {}".format(error_type) if error_type else ""
    full_message = "{} {}{}".format(message, claim_data, error_info)

    # Log the message
    logger = logging.getLogger()
    getattr(logger, level.lower())(full_message)